import logging
import os
from pathlib import Path
from rich.logging import RichHandler

"""
Configures a logger for the current Python script with a rich shell handler for logging.
"""

# create logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# configure shell log
shell_handler = RichHandler()
shell_handler.setLevel(logging.INFO)
fmt_shell = "%(message)s"
shell_formatter = logging.Formatter(fmt_shell)
shell_handler.setFormatter(shell_formatter)
logger.addHandler(shell_handler)



import pickle
from pathlib import Path
import joblib
import json
import numpy as np

import warnings

from typing import Union, List

from xmen.linkers import EntityLinker, logger
from xmen.kb import CompositeKnowledgebase

from scispacy.candidate_generation import (
    CandidateGenerator,
    load_approximate_nearest_neighbours_index,
    LinkerPaths,
    create_tfidf_ann_index,
)
from scispacy.linking import EntityLinker as ScispacyLinker
from scispacy.linking_utils import KnowledgeBase


class TFIDFNGramLinker(EntityLinker):
    """
    A class for linking named entities to concepts in a knowledge base using the TF-IDF and n-gram techniques.

    Args:
    - candidate_generator (CandidateGenerator, optional): The candidate generator object used to retrieve candidate concepts for entity linking. Defaults to None.
    - index_base_path (Union[Path, str], optional): The path to the folder containing the index for the knowledge base. Defaults to None.
    - kb (KnowledgeBase, optional): The knowledge base object used for entity linking. Defaults to None.
    - kb_name (str, optional): The name of the knowledge base. Defaults to "UMLS".
    - expand_abbreviations (bool, optional): Whether to expand abbreviations in entity text. Defaults to True.
    - k (int, optional): The number of candidate concepts to retrieve. Defaults to 30.
    - threshold (float, optional): The similarity threshold for a candidate concept to be considered a match. Defaults to 0.0.
    - no_definition_threshold (float, optional): The similarity threshold for a candidate concept without a definition to be considered a match. Defaults to 0.95.
    - filter_for_definitions (bool, optional): Whether to filter out candidate concepts without a definition. Defaults to False.
    - filter_types (bool, optional): Whether to filter out candidate concepts that are not of type "Entity". Defaults to False.
    """

    @staticmethod
    def write_index(index_base_path: Union[str, Path], jsonl_files: List[Union[str, Path]]):
        """
        Writes the knowledge base to disk in a pickle format and creates an index using the TF-IDF and n-gram techniques.
        - index_base_path: A string or Path object representing the path to the directory where the index will be stored.
        - jsonl_files: A list of string or Path objects representing the paths to the JSONL files containing the knowledge base.
        """
        kb = CompositeKnowledgebase(jsonl_files)

        index_base_path.mkdir(exist_ok=True, parents=True)

        pickle.dump(kb, open(Path(index_base_path) / "kb.pickle", "wb"))

        create_tfidf_ann_index(index_base_path, kb)

    @staticmethod
    def default_scispacy():
        """
        Returns a ScispacyLinker object with default parameters for the candidate generator.
        """
        return TFIDFNGramLinker.scispacy()

    @staticmethod
    def scispacy_no_thresholds():
        """
        Returns a TFIDFNGramLinker object with ScispacyLinker as candidate generator and the given parameters.
        """
        return TFIDFNGramLinker.scispacy(
            k=100,
            threshold=0.0,
            filter_for_definitions=False,
        )

    @staticmethod
    def load_candidate_generator(index_base_path: Union[str, Path]):
        """
        Loads the candidate generator object from the specified index base path.

        Args:
        - index_base_path (Union[str, Path]): A string or Path object representing the base path of the index.

        Returns:
        - CandidateGenerator: A CandidateGenerator object.
        """
        index_base_path = Path(index_base_path)
        lp = LinkerPaths(
            ann_index=index_base_path / "nmslib_index.bin",
            tfidf_vectorizer=index_base_path / "tfidf_vectorizer.joblib",
            tfidf_vectors=index_base_path / "tfidf_vectors_sparse.npz",
            concept_aliases_list=index_base_path / "concept_aliases.json",
        )
        ann_index = load_approximate_nearest_neighbours_index(lp)
        tfidf_vectorizer = joblib.load(lp.tfidf_vectorizer)
        ann_concept_aliases_list = json.load(open(lp.concept_aliases_list, encoding="utf-8"))
        kb = pickle.load(open(index_base_path / "kb.pickle", "rb"))

        cg = CandidateGenerator(ann_index, tfidf_vectorizer, ann_concept_aliases_list, kb)
        return cg

    @staticmethod
    def scispacy(
        expand_abbreviations: bool = None,
        k: int = None,
        threshold: float = None,
        no_definition_threshold: float = None,
        filter_for_definitions: bool = None,
    ):
        """
        Initializes a TFIDFNGramLinker object with a ScispacyLinker candidate generator.

        Args:
        - expand_abbreviations (bool): A boolean indicating whether to expand abbreviations or not. Defaults to None.
        - k (int): An integer indicating the number of candidates to generate for each mention. Defaults to None.
        - threshold (float): A float indicating the similarity threshold required to consider a candidate as a valid match. Defaults to None.
        - no_definition_threshold (float): A float indicating the similarity threshold required to consider a candidate without definition as a valid match. Defaults to None.
        - filter_for_definitions (bool): A boolean indicating whether to filter out candidates without definition or not. Defaults to None.

        Returns:
        - TFIDFNGramLinker: A TFIDFNGramLinker object.
        """
        logger.info("Initializing scispaCy")
        default_linker = ScispacyLinker()
        return TFIDFNGramLinker(
            default_linker.candidate_generator,
            kb=None,
            kb_name="UMLS",
            expand_abbreviations=expand_abbreviations if expand_abbreviations else default_linker.resolve_abbreviations,
            k=k if k else default_linker.k,
            threshold=threshold if threshold else default_linker.threshold,
            no_definition_threshold=no_definition_threshold
            if no_definition_threshold
            else default_linker.no_definition_threshold,
            filter_for_definitions=filter_for_definitions
            if filter_for_definitions
            else default_linker.filter_for_definitions,
            filter_types=False,
        )

    # Ignore caching when using dataset.map
    def __getstate__(self):
        return {}

    def __init__(
        self,
        candidate_generator: CandidateGenerator = None,
        index_base_path: Union[Path, str] = None,
        kb: KnowledgeBase = None,
        kb_name: str = "UMLS",
        expand_abbreviations: bool = True,
        k: int = 30,
        threshold: float = 0.0,
        no_definition_threshold: float = 0.95,
        filter_for_definitions: bool = False,
        filter_types: bool = False,
    ):
        if candidate_generator:
            self.candidate_generator = candidate_generator
        elif index_base_path:
            self.candidate_generator = TFIDFNGramLinker.load_candidate_generator(index_base_path)
        else:
            raise Exception("Either candidate_generator or index_base_path have to be provided")
        self.expand_abbreviations = expand_abbreviations
        self.k = k
        self.kb = kb if kb else self.candidate_generator.kb
        self.kb_name = kb_name
        self.threshold = threshold
        self.no_definition_threshold = no_definition_threshold
        self.filter_for_definitions = filter_for_definitions
        self.filter_types = filter_types

    def predict(self, passages: list, entities: list) -> dict:
        """
        Predicts the normalized form of the entities based on the provided passages and returns the updated entities object.

        Args:
        - passages (list): A list of strings representing the passages where the entities are found.
        - entities (list): A list of dictionaries representing the entities. Each dictionary must contain at least the "text" key with the entity text.

        Returns:
        - entities: A dictionary representing the updated entities object.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", np.VisibleDeprecationWarning)

            def get_m_string(e):
                if self.expand_abbreviations and e.get("long_form", None):
                    return e["long_form"]
                else:
                    return " ".join(e["text"])

            mention_strings = [get_m_string(e) for e in entities]
            batch_candidates = self.candidate_generator(mention_strings, self.k)

            for _, candidates, e in zip(mention_strings, batch_candidates, entities):
                predicted = []
                for cand in candidates:
                    score = max(cand.similarities)
                    if (
                        self.filter_for_definitions
                        and self.kb.cui_to_entity[cand.concept_id].definition is None
                        and score < self.no_definition_threshold
                    ):
                        continue
                    if score > self.threshold:
                        predicted.append((cand.concept_id, score))
                sorted_predicted = sorted(predicted, reverse=True, key=lambda x: x[1])
                e["normalized"] = [
                    {"db_id": cui, "score": score, "db_name": self.kb_name, "predicted_by": ["ngram"]}
                    for cui, score in sorted_predicted
                ]
        return entities



from xmen.umls import get_sem_type_tree, expand_tuis
from typing import Dict


class SemanticTypeFilter:
    """
    A class to filter out examples based on semantic types.

    Args:
    - type_to_tui (Dict): A dictionary that maps a semantic type to a set of TUIs.
    - kb: A knowledge base object.
    - expand_types: Whether to expand the given TUIs based on the semantic type tree

    Attributes:
    - type_to_tui (Dict): A dictionary that maps a semantic type to a set of TUIs.
    - kb: A knowledge base object.
    - tree: A semantic type tree object.
    - expand_types: flag indicating whether to expand semantic types
    """

    def __init__(self, type_to_tui: Dict, kb, expand_types=False):
        self.type_to_tui = type_to_tui
        self.kb = kb
        self.tree = get_sem_type_tree()
        self.expand_types = expand_types

    def __getstate__(self):
        return {}

    def get_tuis(self, cui):
        """
        Returns the set of TUIs associated with a given CUI.

        Args:
        - cui (str): A CUI string.

        Returns:
        - tuis (set): A set of TUIs associated with the given CUI.
        """
        return self.kb.cui_to_entity[cui].types

    def filter_semantic_types(self, example):
        """
        Filters out normalized entities from the given example that are not associated with any of the valid TUIs.

        Args:
        - example (dict): A dictionary representing a single example, with keys "text" and "entities".

        Returns:
        - filtered_example (dict): A dictionary representing the filtered example, with key "entities".
        """
        entities = example["entities"]
        for e in entities:
            valid_tuis = self.type_to_tui[e["type"]]
            if self.expand_types:
                valid_tuis = expand_tuis(valid_tuis, self.tree)
            filtered = []
            for n in e["normalized"]:
                concept_tuis = self.get_tuis(n["db_id"])
                if any([t for t in concept_tuis if t in valid_tuis]):
                    filtered.append(n)
            e["normalized"] = filtered
        return {"entities": entities}

    def transform_batch(self, ds):
        """
        Transforms the given dataset by applying the filter_semantic_groups method to each example in the dataset.

        Args:
        - ds (tf.data.Dataset): A dataset of examples.

        Returns:
        - transformed_ds (tf.data.Dataset): A transformed dataset of examples.
        """
        return ds.map(lambda e: self.filter_semantic_types(e), load_from_cache_file=False)



class EmptyNormalizationFilter:
    """
    A class to filter out entities with empty normalized values from text data.
    """

    def transform_batch(self, dataset):
        """
        Filters entities with empty normalized values from a given dataset.

        Args:
        - dataset (Dataset): A dataset object from the Hugging Face datasets library.

        Returns:
        - Dataset: A transformed dataset object with the same number of examples, but with empty-normalized entities removed.
        """

        def filter_empty(d):
            ents = [e for e in d["entities"] if len(e["normalized"]) > 0]
            return {"entities": ents}

        return dataset.map(filter_empty)


class MissingCUIFilter:
    """
    A class to filter out entities with missing CUIs according to a given knowledge base.
    """

    def __init__(self, kb):
        self.kb = kb

    def __getstate__(self):
        return {}

    def transform_batch(self, dataset):
        def filter_missing_cuis(entities):
            result = []
            for e in entities:
                filtered = []
                for n in e["normalized"]:
                    if n["db_id"] in self.kb.cui_to_entity:
                        filtered.append(n)
                if len(filtered) > 0:
                    e["normalized"] = filtered
                    result.append(e)
            return result

        return dataset.map(lambda i: {"entities": filter_missing_cuis(i["entities"])})



#!/usr/bin/env python
"Document - group and compare annotations"
"Adapted from https://github.com/wikilinks/neleval/blob/master/neleval/ for Python >3.10 compatibility"

from collections import OrderedDict
import warnings
import sys
import argparse

from .annotation import Annotation
from neleval.utils import unicode

TEMPLATE = u'{}\t{}\t{}\t{}\t{}'
ENC = 'utf8'


# Document class contains methods for linking annotation

class Document(object):
    def __init__(self, id, annotations):
        self.id = id
        self.annotations = sorted(annotations, key=lambda a: (a.start, -a.end))
        self._validate()
        self._set_fields()

    # May be 'ignore', 'warn', 'error'
    VALIDATION = {
        'nested': 'ignore',
        'crossing': 'ignore',
        'duplicate': 'ignore',
    }

    def _validate(self, _categories=['nested', 'crossing', 'duplicate']):
        # XXX: do we nee to ensure start > end for all Annotations first?
        issues = {cat: [] for cat, val in self.VALIDATION.items()
                  if val != 'ignore'}
        if not issues:
            return
        open_anns = []
        tags = sorted([(a.start, 'open', a) for a in self.annotations] +
                      [(a.end + 1, 'close', a) for a in self.annotations])  # use stop, not end
        for key, op, ann in tags:
            if op == 'open':
                open_anns.append(ann)
            else:
                open_anns.remove(ann)
                for other in open_anns:
                    comparison = ann.compare_spans(other)
                    if comparison in issues:
                        issues[comparison].append((other, ann))
        for issue, instances in issues.items():
            if not instances:
                continue
            if self.VALIDATION[issue] == 'error':
                b, a = instances[0]
                raise ValueError(u'Found annotations with {} span:\n{}\n{}'
                                 .format(issue, unicode(a), unicode(b)))
            elif self.VALIDATION[issue] == 'warn':
                b, a = instances[0]
                warnings.warn(u'Found annotations with {} span:\n{}\n{}'
                              .format(issue, unicode(a), unicode(b)))

    def _set_fields(self):
        """Set fields on annotations that are relative to document"""
        seen = set()
        for a in self.annotations:
            eid = a.eid
            a.is_first = eid not in seen
            seen.add(eid)

    def __unicode__(self):
        return u'\n'.join(unicode(a) for a in self.annotations)

    __str__ = __unicode__

    # Accessing Spans.
    def _iter_mentions(self, link=True, nil=True):
        assert not (not link and not nil), 'Must filter some mentions.'
        for a in self.annotations:
            #TODO check logic, handle TAC NILs
            if not link and a.is_linked:
                continue # filter linked mentions
            if not nil and a.is_nil:
                continue # filter nil mentions
            yield a

    def iter_mentions(self):
        return self._iter_mentions(link=True, nil=True)

    def iter_links(self):
        return self._iter_mentions(link=True, nil=False)

    def iter_nils(self):
        return self._iter_mentions(link=False, nil=True)


# Grouping annotations

def by_document(annotations):
    d = OrderedDict()
    for a in annotations:
        if a.docid in d:
            d[a.docid].append(a)
        else:
            d[a.docid] = [a]
    return d.items()


def by_mention(annotations):
    return [("{}//{}..{}".format(a.docid, a.start, a.end), [a])
            for a in annotations]


# Reading annotations

class Reader(object):
    "Read annotations, grouped into documents"
    def __init__(self, fh, group=by_document, cls=Document):
        self.fh = fh
        self.group = group
        self.cls = cls

    def __iter__(self):
        return self.read()

    def read(self):
        try:
            for groupid, annots in self.group(self.annotations()):
                yield self.cls(groupid, annots)
        except Exception:
            print('ERROR while processing', self.fh, file=sys.stderr)
            raise

    def annotations(self):
        "Yield Annotation objects"
        for line in self.fh:
            yield Annotation.from_string(line.rstrip('\n'))


class ValidateSpans(object):
    """Identify duplicate, crossing and nested spans

    Will output warnings or errors as determined by options.
    """
    def __init__(self, system, duplicate='error', crossing='warn', nested='ignore'):
        self.system = system
        self.duplicate = duplicate
        self.crossing = crossing
        self.nested = nested

    def __call__(self):
        # HACK (not concurrency-safe)
        OLD_VALIDATION = Document.VALIDATION
        Document.VALIDATION = {k: getattr(self, k)
                               for k in OLD_VALIDATION}
        list(Reader(self.system))
        Document.VALIDATION = OLD_VALIDATION

    @classmethod
    def add_arguments(cls, p):
        CHOICES = ['ignore', 'warn', 'error']
        p.add_argument('system', nargs='?', default=sys.stdin, type=argparse.FileType('r'), metavar='FILE')
        p.add_argument('--duplicate', default='error', choices=CHOICES)
        p.add_argument('--crossing', default='warn', choices=CHOICES)
        p.add_argument('--nested', default='ignore', choices=CHOICES)
        p.set_defaults(cls=cls)
        return p



from .indexed_dataset import IndexedDatasetDict, IndexedDataset
from .util import *
from .integrations import *

from .abbrevations import AbbreviationExpander
from .sampling import Sampler
from .merge_concepts import ConceptMerger
from .retired_cuis import CUIReplacer
from .filter import EmptyNormalizationFilter, MissingCUIFilter
from .semantic_types import SemanticTypeFilter
from .semantic_groups import SemanticGroupFilter
from .deduplication import Deduplicator

from datasets import Features, Value, Sequence

features = Features(
    {
        "id": Value(dtype="string"),
        "document_id": Value(dtype="string"),
        "passages": [
            {
                "id": Value(dtype="string"),
                "type": Value(dtype="string"),
                "text": Sequence(feature=Value(dtype="string")),
                "offsets": Sequence(feature=[Value(dtype="int32")]),
            }
        ],
        "entities": [
            {
                "id": Value(dtype="string"),
                "offsets": Sequence(feature=Sequence(feature=Value(dtype="int32"))),
                "text": Sequence(feature=Value(dtype="string")),
                "type": Value(dtype="string"),
                "normalized": [
                    {
                        "db_id": Value(dtype="string"),
                        "db_name": Value(dtype="string"),
                        "score": Value(dtype="float"),
                        "predicted_by": Sequence(Value(dtype="string")),
                    }
                ],
                "long_form": Value(dtype="string"),
            }
        ],
        "events": [
            {
                "id": Value(dtype="string"),
                "type": Value(dtype="string"),
                "trigger": {
                    "text": Sequence(feature=Value(dtype="string")),
                    "offsets": Sequence(feature=[Value(dtype="int32")]),
                },
                "arguments": [{"role": Value(dtype="string"), "ref_id": Value(dtype="string")}],
            }
        ],
        "coreferences": [
            {
                "id": Value(dtype="string"),
                "entity_ids": Sequence(feature=Value(dtype="string")),
            }
        ],
        "relations": [
            {
                "id": Value(dtype="string"),
                "type": Value(dtype="string"),
                "arg1_id": Value(dtype="string"),
                "arg2_id": Value(dtype="string"),
                "normalized": [{"db_name": Value(dtype="string"), "db_id": Value(dtype="string")}],
            }
        ],
        "corpus_id": Value(dtype="string"),
        "lang": Value(dtype="string"),
    }
)



import os
import logging
import typer
from ..log import logger
from pathlib import Path
from ..confhelper import load_config

cfg_path_help = "Path to the .yaml file specifying the configuration."
code_help = "Path to the python script to parse a non-umls kwnoledge base."
output_help = (
    "Path to the directory where the files will be generated. If not provided, the default will be "
    "~/.cache/xmen/{kb_name}."
)

log_file_help = (
    "Path to store the log output locally. Multiple commands to the same log will be "
    "appended, not overwritten. If not provided, no log will be stored."
)

dict_dir_help = (
    "Custom path to the .jsonl dict file from which the indicess will be generated. Provide only if such "
    ".jsonl is not stored in {work_dir} in the config .yaml file."
)

overwrite_help = (
    "If set to True, command will by default overwrite dicts and indexes when encountered with the same "
    "output directories and files, instead of asking if the user wishes to overwrite for each case."
)

custom_name_help = (
    "Provide a name for the dictionary to be built other than the name specified in the config .yaml"
    ". Use for datasets with multiple knowledge bases. The file extension is added automatically."
)

key_help = "Provide the name of the key in the .yaml config file that contains the desired subconfig for" " the dict."

sapbert_help = "Build SapBERT indices."
ngram_help = "Build N-Gram indices."
all_help = "Build all available indicess"
gpu_id_help = "Indicate in which cuda GPU device to generate the indices. Use `-1` to run on the CPU instead."
batch_help = "GPU batch size to use for SapBERT"
index_buffer_size_help = "Buffer size for use for writing SapBERT FAISS Index"
save_ram_help = "Write temporary files to disk to save RAM for SapBERT indices."


def get_work_dir(cfg) -> Path:
    """
    Retrieves the working directory specified in the configuration or creates a default directory.

    Args:
    - cfg: a dictionary containing configuration parameters

    Returns:
    - Path object representing the working directory
    """
    if "work_dir" in cfg:
        return Path(cfg.work_dir)
    else:  # default build path
        return Path(os.path.expanduser("~")) / ".cache/xmen" / f"{cfg.name}"


def add_file_logger(log_file):
    """
    Adds a file handler to the logger to save log output to a file.

    Args:
    - log_file: a string representing the path to the log file

    Returns:
    - None
    """
    file_handler = logging.FileHandler(Path(log_file))
    file_handler.setLevel(logging.DEBUG)
    fmt_file = "%(asctime)s - %(levelname)s - %(message)s"
    file_formatter = logging.Formatter(fmt_file)
    file_handler.setFormatter(file_formatter)
    logger.addHandler(file_handler)
    pass


def check_and_load_config(cfg_path):
    """
    Warns if the provided path to the configuration file is invalid, otherwise loads it as a DictConf object.

    Args:
    - cfg_path (str): Path to the configuration file to be loaded.

    Returns:
    - (DictConf): Loaded configuration file as a DictConf object.

    Raises:
    - typer.Exit: If the provided path does not exist or is not in the .yaml format.
    """
    if not os.path.exists(cfg_path) or not cfg_path.endswith(".yaml"):
        logger.error(f"File {cfg_path} does not exist or is not in the supported .yaml format.")
        raise typer.Exit()
    else:
        return load_config(cfg_path)


def can_write(path: str, parent_overwrite: bool) -> bool:
    """
    Checks whether the files to be built (dicts or indices) already exist. If so, gives the chance to overwrite them in case it is not prohibited by parent_overwrite flag. the function prompts the user to input either 'y' or 'n' if the file already exists.

    Args:
    - file_path (str): Path to the file being checked.
    - parent_overwrite (bool): Whether the parent command has the `--overwrite` flag set to True.

    Returns:
    - (bool): True if the file can be overwritten, False otherwise.
    """
    if parent_overwrite:
        return True
    else:
        # Handle differently file and directory paths (for dict and index, respectively)
        if "." in str(path).split("/")[-1]:  # check if it is a file, os.path.isfile() works only if the path exists
            is_warnable = os.path.exists(path)
            warning = f"There already exists a file in {path}."
        else:
            is_warnable = os.path.exists(path) and os.listdir(path)
            warning = f"There are already some files in {path}. This command may overwrite them."

        if is_warnable:
            logger.warn(warning)
            logger.info(
                f"To instruct the command to overwrite all coinciding files by default, switch the flag --overwrite."
                f"You can also provide a different path to write the files with --output. \n"
                f"What would you like to do this time? Input yes [y] to overwrite it, no [n] to skip the file."
            )
            overwrite = input("[y/n]")
        else:
            overwrite = "y"

        if overwrite not in ["n", "y"]:
            logger.warn("Unhandled value. Please type `y` for yes or `n` for no.")
            overwrite = can_write(path, parent_overwrite)

        return True if overwrite == "y" else False


def get_alias_count(concept_details):
    """
    Calculates the total number of aliases in the given concept_details dictionary.

    Args:
    - concept_details (dict): A dictionary containing details of concepts and their aliases.

    Returns:
    - (int): Total number of aliases, including the original concept name itself.
    """
    return sum([len(c["aliases"]) + 1 for c in concept_details.values()])



from xmen.ext.scispacy.umls_utils import read_umls_file_headers


class CUIReplacer:
    """
    A class that replaces retired CUIs in normalized entities with their replacements.

    Args:
    - umls_meta_path (str): The path to the UMLS metadata directory

    Attributes:
    - retired_cuis (dict): A dictionary containing retired CUIs as keys and their replacements as values
    """

    def __init__(self, umls_meta_path):
        self.retired_cuis = self._load_retired_cuis(umls_meta_path)

    def transform_batch(self, dataset):
        """
        Replaces retired CUIs in the normalized entities of a dataset with their replacements.

        Args:
        - dataset (MapDataset): A dataset containing entities with normalized forms

        Returns:
        - MapDataset: The transformed dataset with replaced CUIs in the normalized entities
        """

        def replace_cuis(entities):
            """
            Helper function that replaces retired CUIs in normalized forms of entities with their replacements.

            Args:
            - entities (list): A list of entities containing normalized forms with CUIs to replace

            Returns:
            - list: A list of entities with replaced CUIs in the normalized forms
            """
            for e in entities:
                for n in e["normalized"]:
                    c = n["db_id"]
                    n["db_id"] = self.retired_cuis.get(c, c)
            return entities

        return dataset.map(lambda i: {"entities": replace_cuis(i["entities"])})

    def _load_retired_cuis(self, meta_path):
        """
        Loads the retired CUIs mapping from the MRCUI.RRF file in the metadata directory.

        Args:
        - meta_path (str): The path to the UMLS metadata directory

        Returns:
        - dict: A dictionary containing retired CUIs as keys and their replacements as values
        """
        res = []
        cui_filename = "MRCUI.RRF"
        headers = read_umls_file_headers(meta_path, cui_filename)
        mapping = {}
        with open(f"{meta_path}/{cui_filename}") as fin:
            for line in fin:
                splits = line.strip().split("|")
                assert len(headers) == len(splits)
                res = dict(zip(headers, splits))
                if res["CUI2"]:
                    mapping[res["CUI1"]] = res["CUI2"]
        return mapping



import pandas as pd
from tqdm import tqdm
from xmen.evaluation import evaluate, entity_linking_error_analysis
import matplotlib
import matplotlib.pyplot as plt


def eval_thresholds(dataset, pred, thresholds, ks, pbar=None):
    """
    Evaluates a set of predictions for different threshold values and top-k metrics.

    Args:
    - dataset (Iterable): Iterable of dictionaries containing the ground truth entities.
    - pred (Iterable): Iterable of dictionaries containing the predicted entities.
    - thresholds (Iterable): Iterable of float values representing the different threshold values to evaluate.
    - ks (Iterable[int]): Iterable of integer values representing the different top-k metrics to evaluate.
    - pbar (Optional[TqdmProgressBar]): A progress bar object to track the evaluation progress. Default is None.

    Returns:
    A dictionary containing the evaluation metrics for each threshold value and top-k metric.
    """
    all_eval = {}

    steps = len(thresholds) * len(ks)

    if not pbar:
        pbar = tqdm(total=steps)
    for threshold in thresholds:
        all_eval[threshold] = {}
        for k in ks:
            eval = evaluate(
                dataset,
                pred,
                allow_multiple_gold_candidates=False,
                top_k_predictions=k,
                threshold=threshold,
            )
            all_eval[threshold][f"f1@{k}"] = eval["strict"]["fscore"]
            all_eval[threshold][f"p@{k}"] = eval["strict"]["precision"]
            all_eval[threshold][f"r@{k}"] = eval["strict"]["recall"]
            pbar.update(1)
    return all_eval


def run_eval(dataset, preds, ks: list, thresholds: list = None):
    """
    Computes evaluation metrics and entity linking error analysis for a set of predictions.

    Args:
    - dataset (Iterable): Iterable of dictionaries containing the ground truth entities.
    - pred (Iterable): Iterable of dictionaries containing the predicted entities.
    - ks (Iterable[int]): Iterable of integer values representing the different top-k metrics to evaluate.
    - thresholds (Iterable): Iterable of float values representing the different threshold values to evaluate. Defaults to None.

    Returns:
    - evals: a list of dictionaries containing evaluation metrics (e.g., precision, recall, F1)
    - error_dfs: a list of error dataframes, one for each prediction, containing entity linking error analysis
    """
    evals = []
    error_dfs = []

    if thresholds:
        steps = len(thresholds) * len(ks) * len(preds)
    else:
        steps = len(preds)

    with tqdm(total=steps) as pbar:
        for pred in preds:
            if thresholds:
                all_eval = eval_thresholds(dataset, pred, thresholds, ks, pbar)
                evals.append(all_eval)

            error_df = entity_linking_error_analysis(dataset, pred)
            error_dfs.append(error_df)
    return evals, error_dfs


def merge_eval_df(linker_names, evals):
    """
    Merges evaluation metrics into a single dataframe.

    Args:
    - linker_names: a list of strings representing the names of the entity linkers
    - evals: a list of dictionaries containing evaluation metrics for each linker and threshold

    Returns:
    - eval_df: a pandas dataframe with columns representing the evaluation metrics and index
      containing a MultiIndex of linker names and threshold values
    """
    eval_df = []
    for e, l in zip(evals, linker_names):
        for t, res in e.items():
            res["threshold"] = t
            res["linker"] = l
            eval_df.append(res)
    eval_df = pd.DataFrame(eval_df)
    eval_df.set_index(["linker", "threshold"], inplace=True)
    return eval_df


def plot_eval_results_at_k(eval_df, k_range, eval_names):
    """
    Plots evaluation results at various values of k for a set of entity linkers.

    Args:
    - eval_df: a pandas dataframe containing evaluation metrics for each entity linker and threshold
    - k_range: a list of integers representing the cutoffs for precision@k and recall@k
    - eval_names: a list of strings representing the names of the entity linkers

    Returns:
    - None
    """
    # matplotlib.rcParams.update({'font.size': 20})

    fig, axs = plt.subplots(1, 3, figsize=(22, 6))

    for j, key in enumerate(["p", "r", "f1"]):
        ax = axs[j]
        cols = [f"{key}@{i}" for i in k_range]
        subset_df = eval_df[cols]
        subset_df.columns = k_range
        subset_df.T.plot(ax=ax, legend=False, style=[".--", "s:", "x-", "x--", "o-"])
        ax.grid(color="lightgray", linestyle="-", linewidth=0.5)
        ax.set_ylim(0.0, 0.8)
        ax.set_xticks(k_range)
        ax_labels = {"p": "Precision", "r": "Recall", "f1": "F1"}
        ax.set_title(ax_labels[key])
        ax.set_xlabel("k")

    handles, labels = axs[0].get_legend_handles_labels()

    fig.legend(
        handles,
        eval_names,
        loc="center",
        ncol=3,
        bbox_to_anchor=(0.512, -0.1, 0.00, 0.0),
    )

    plt.subplots_adjust(wspace=0.20)

    plt.xticks(k_range)



from transformers import AutoModelForSeq2SeqLM, AutoTokenizer
import torch
import spacy
import re
from xmen.log import logger
from datasets import DatasetDict
from typing import List, Iterator


class Translator:

    ENGLISH = "eng_Latn"
    GERMAN = "deu_Latn"
    FRENCH = "fra_Latn"
    SPANISH = "spa_Latn"
    DUTCH = "nld_Latn"

    def __init__(self, src_lang: str, target_lang: str, cuda: bool = True, batch_size: int = 4) -> None:
        self.src_lang = src_lang
        self.target_lang = target_lang
        self.batch_size = batch_size

        logger.info("Initializing Translation Model")

        self.tokenizer = AutoTokenizer.from_pretrained("facebook/nllb-200-distilled-600M", src_lang=src_lang)

        self.model = AutoModelForSeq2SeqLM.from_pretrained("ychenNLP/nllb-200-3.3b-easyproject")
        if cuda:
            self.model.cuda()

        self.nlp = spacy.load("en_core_sci_sm")

    def __getstate__(self):
        return {}

    def _annos_to_tagged(self, document):
        tagged_passages = []
        e_id = 0
        for p in document["passages"]:
            tagged_passage = {"id": p["id"], "type": p["type"], "text": []}
            for text, passage_offset in zip(p["text"], p["offsets"]):
                for sent in self.nlp(text).sents:
                    o = (passage_offset[0] + sent.start_char, passage_offset[0] + sent.end_char)
                    entities = [
                        e for e in document["entities"] if e["offsets"][0][0] >= o[0] and e["offsets"][-1][1] <= o[1]
                    ]
                    transformed_text = sent.text
                    offset_diff = -o[0]

                    for entity in entities:
                        e_id += 1
                        start = entity["offsets"][0][0] + offset_diff
                        end = entity["offsets"][-1][1] + offset_diff
                        s_start = f"[{e_id}]"
                        s_end = f"[/{e_id}]"
                        transformed_text = (
                            transformed_text[:start]
                            + s_start
                            + transformed_text[start:end]
                            + s_end
                            + transformed_text[end:]
                        )
                        offset_diff += len(str(e_id)) * 2 + 5  # Length of entity_type + 5 (for the tags)
                    tagged_passage["text"].append(transformed_text)
            tagged_passages.append(tagged_passage)
        return {"passages": tagged_passages}

    def _translate_tagged_document(self, document):
        output_passages = []

        for p in document["passages"]:
            translated_passage = p.copy()
            tagged_sentences = p["text"]
            translated_sentences = []
            for idx in range(0, len(tagged_sentences), self.batch_size):
                start_idx = idx
                end_idx = idx + self.batch_size
                inputs = self.tokenizer(
                    tagged_sentences[start_idx:end_idx],
                    padding=True,
                    truncation=True,
                    max_length=512,
                    return_tensors="pt",
                ).to("cuda")
                max_input_length = max([len(i) for i in inputs["input_ids"]])

                with torch.no_grad():
                    translated_tokens = self.model.generate(
                        **inputs,
                        forced_bos_token_id=self.tokenizer.lang_code_to_id[self.target_lang],
                        max_length=max(128, int(max_input_length * 1.5)),
                        num_beams=5,
                        num_return_sequences=1,
                        early_stopping=True,
                    )

                output = self.tokenizer.batch_decode(translated_tokens, skip_special_tokens=True)
                translated_sentences.extend(output)
            translated_passage["text"] = translated_sentences
            output_passages.append(translated_passage)
        return {"passages": output_passages}

    def _remove_nested_tags_text(self, t, document_id):
        pattern = "\[(\/)?(\d+)\]"
        clean_text = t
        opened = []
        index = 0
        while m := re.search(pattern, clean_text[index:]):
            index += m.span()[1]
            if not m.group(1):  # is start tag
                m_next = re.search(pattern, clean_text[index:])
                if not m.group(2) in opened and m_next and m_next.group(1) == "/" and m.group(2) == m_next.group(2):
                    opened.append(m.group(2))
                    continue  # alright
                else:  # Invalid tag
                    clean_text = clean_text.replace(m.group(0), "", 1)
                    index -= len(m.group(0))
                    if not m.group(0) in clean_text:  # again
                        logger.debug(f"Document {document_id} - Replacing nested tag {m.group(0)} in:\n{t}")
                        clean_text = clean_text.replace(f"[/{m.group(2)}]", "")
                    else:
                        logger.debug(f"Document {document_id} - Replacing single tag {m.group(0)} in:\n{t}")
            else:  # is end tag
                if not m.group(2) in opened:
                    logger.debug(f"Document {document_id} - Closing unopened tag {m.group(0)} in: \n{t}")
                    clean_text = clean_text.replace(m.group(0), "", 1)
                    index -= len(m.group(0))
        return clean_text

    def _remove_nested_entities(self, document):
        passages = document["passages"].copy()
        for p in passages:
            clean_texts = []
            for t in p["text"]:
                errors = True
                while errors:
                    clean_text = self._remove_nested_tags_text(t, document["id"])
                    if t == clean_text:
                        logger.debug("Cleaned document:\n" + clean_text)
                        clean_texts.append(clean_text)
                        errors = False
                    t = clean_text
            p["text"] = clean_texts
        return {"passages": passages}

    def _realign_entities(self, document):
        eid2entity = {str(i + 1): e for i, e in enumerate(document["entities"])}
        passages = []
        entities = []
        start_offset = 0
        for p in document["passages"]:
            new_passage = p.copy()
            new_passage["text"] = []
            new_passage["offsets"] = []
            for t in p["text"]:
                orig_text = t
                sentence_entities = []
                while start_match := re.search("\[(\d+)\]", t):
                    start_span = start_match.span()
                    e_id = start_match.group(1)
                    other_start_match = re.findall(f"\[({e_id})\]", t)
                    end_match = re.search(f"\[\/({e_id})\]", t)
                    if len(other_start_match) > 1 or not end_match:
                        logger.debug(f"Unclosed tag {start_match} in {orig_text} - Skipping")
                        t = t[: start_span[0]] + t[start_span[1] :]
                        continue

                    end_span = end_match.span()
                    ent_text = t[start_span[1] : end_span[0]]
                    t = t[: start_span[0]] + ent_text + t[end_span[1] :]
                    if not e_id in eid2entity:
                        logger.debug(f"{e_id} does not match original entities - Skipping")
                    elif not ent_text.strip():
                        logger.debug(f"Empty entity {e_id} - Skipping")
                    else:
                        entity = eid2entity[e_id].copy()
                        entity["text"] = [ent_text]
                        entity["offsets"] = [
                            [start_offset + start_span[0], start_offset + end_span[0] - len(start_match.group(0))]
                        ]
                        entities.append(entity)
                for e in sentence_entities:
                    assert t[e["offset"][0] : e["offset"][1]] == e["text"], (e, t)
                new_passage["text"].append(t)
                new_passage["offsets"].append([start_offset, start_offset + len(t)])
                start_offset += len(t) + 1
            passages.append(new_passage)
        found_ids = {e["id"] for e in entities}
        missing_ids = []
        for e in eid2entity.values():
            if not e["id"] in found_ids:
                missing_ids.append(e["id"])
        if missing_ids:
            logger.debug(f"Entities {missing_ids} missing from translation in document {document['id']}")
        return {"passages": passages, "entities": entities}

    def _translate_batch(self, dataset):
        tagged_dataset = dataset.map(self._annos_to_tagged, load_from_cache_file=False)
        logger.info("Translating documents")
        return tagged_dataset.map(self._translate_tagged_document, load_from_cache_file=False)

    def _tags_to_entities(self, dataset):
        clean_tagged_dataset = dataset.map(self._remove_nested_entities, load_from_cache_file=False)
        realigned_dataset = clean_tagged_dataset.map(self._realign_entities, load_from_cache_file=False)
        check_passages_offsets(realigned_dataset)
        check_entities_offsets(realigned_dataset)
        return realigned_dataset

    def transform_batch(self, dataset):
        translated_tagged_dataset = self._translate_batch(dataset)
        return self._tags_to_entities(translated_tagged_dataset)


# Offset checks adapted from https://github.com/bigscience-workshop/biomedical/blob/main/tests/test_bigbio.py


def _get_example_text(example: dict) -> str:
    """
    Concatenate all text from passages in an example of a KB schema
    :param example: An instance of the KB schema
    """
    return " ".join([t for p in example["passages"] for t in p["text"]])


def check_passages_offsets(dataset_bigbio: DatasetDict):
    for split in dataset_bigbio:
        if "passages" in dataset_bigbio[split].features:
            for example in dataset_bigbio[split]:
                example_text = _get_example_text(example)

                for passage in example["passages"]:
                    example_id = example["id"]
                    text = passage["text"]
                    offsets = passage["offsets"]
                    for idx, (start, end) in enumerate(offsets):
                        msg = (
                            f"Split:{split} - Example:{example_id} - "
                            f"text:`{example_text[start:end]}` != text_by_offset:`{text[idx]}`"
                        )
                        if not example_text[start:end] == text[idx]:
                            logger.error(msg)


def _check_offsets(
    example_id: int,
    split: str,
    example_text: str,
    offsets: List[List[int]],
    texts: List[str],
) -> Iterator:
    if len(texts) != len(offsets):
        logger.warning(
            f"Split:{split} - Example:{example_id} - "
            f"Number of texts {len(texts)} != number of offsets {len(offsets)}. "
            f"Please make sure that this error already exists in the original "
            f"data and was not introduced in the data loader."
        )
    # offsets are always list of lists
    for idx, (start, end) in enumerate(offsets):
        by_offset_text = example_text[start:end]
        try:
            text = texts[idx]
        except IndexError:
            text = ""
        if by_offset_text != text:
            yield f" text:`{text}` != text_by_offset:`{by_offset_text}`"


def check_entities_offsets(dataset_bigbio: DatasetDict):
    errors = []
    for split in dataset_bigbio:
        if "entities" in dataset_bigbio[split].features:
            for example in dataset_bigbio[split]:
                example_id = example["id"]
                example_text = _get_example_text(example)

                for entity in example["entities"]:
                    for msg in _check_offsets(
                        example_id=example_id,
                        split=split,
                        example_text=example_text,
                        offsets=entity["offsets"],
                        texts=entity["text"],
                    ):
                        entity_id = entity["id"]
                        errors.append(f"Split: {split}, Example:{example_id} - entity:{entity_id} " + msg)
    if len(errors) > 0:
        logger.warning(msg="\n".join(errors) + " Wrong offsets")



"Adapted from https://github.com/wikilinks/neleval/blob/master/neleval/ for Python >3.10 compatibility"

import textwrap
from collections import defaultdict
from .annotation import Measure, Annotation

try:
    keys = dict.viewkeys
except Exception:
    # Py3k
    keys = dict.keys


MEASURES = {
    # Mention evaluation measures
    'strong_mention_match':         Measure(['span']),
    'strong_typed_mention_match':   Measure(['span', 'type']),
    'strong_linked_mention_match':  Measure(['span'], 'is_linked'),
    # Linking evaluation measures
    'strong_link_match':            Measure(['span', 'kbid'], 'is_linked'),
    'strong_nil_match':             Measure(['span'], 'is_nil'),
    'strong_all_match':             Measure(['span', 'kbid']),
    'strong_typed_link_match':      Measure(['span', 'type', 'kbid'],
                                            'is_linked'),
    'strong_typed_nil_match':       Measure(['span', 'type'], 'is_nil'),
    'strong_typed_all_match':       Measure(['span', 'type', 'kbid']),
    # Document-level tagging evaluation measures
    'entity_match':                 Measure(['docid', 'kbid'], 'is_linked'),
    # Clustering evaluation measures
    'muc':                          Measure(['span'], agg='muc'),
    'b_cubed':                      Measure(['span'], agg='b_cubed'),
    'b_cubed_plus':                 Measure(['span', 'kbid'], agg='b_cubed'),
    'entity_ceaf':                  Measure(['span'], agg='entity_ceaf'),
    'mention_ceaf':                 Measure(['span'], agg='mention_ceaf'),
    'mention_ceaf_plus':            Measure(['span', 'kbid'], agg='mention_ceaf'),
    'typed_mention_ceaf':           Measure(['span', 'type'], agg='mention_ceaf'),
    'typed_mention_ceaf_plus':      Measure(['span', 'type', 'kbid',], agg='mention_ceaf'),
    'pairwise':                     Measure(['span'], agg='pairwise'),
    # Cai & Strube (2010) evaluation measures
    #'cs_b_cubed':                   Measure(['span'], agg='cs_b_cubed'),
    #'entity_cs_ceaf':               Measure(['span'], agg='entity_cs_ceaf'),
    #'mention_cs_ceaf':              Measure(['span'], agg='mention_cs_ceaf'),
}


# Configuration constants
ALL_MEASURES = 'all'
ALL_TAGGING = 'all-tagging'
ALL_COREF = 'all-coref'
TAC09_MEASURES = 'tac09'   # used 2009-2010
TAC11_MEASURES = 'tac11'   # used 2011-2013
TAC14_MEASURES = 'tac14'   # used 2014-
TMP_MEASURES = 'tmp'
CORNOLTI_WWW13_MEASURES = 'cornolti'
HACHEY_ACL14_MEASURES = 'hachey'
LUO_MEASURES = 'luo'
CAI_STRUBE_MEASURES = 'cai'

MEASURE_SETS = {
    ALL_MEASURES: [
        ALL_TAGGING,
        ALL_COREF,
        ],
    ALL_TAGGING: {
        'strong_mention_match',
        'strong_typed_mention_match',
        'strong_linked_mention_match',
        'strong_link_match',
        'strong_nil_match',
        'strong_all_match',
        'strong_typed_link_match',
        'strong_typed_nil_match',
        'strong_typed_all_match',
        'entity_match',
    },
    ALL_COREF: {
        'mention_ceaf',
        'entity_ceaf',
        'b_cubed',
        'pairwise',
        'muc',
        #'mention_cs_ceaf',
        #'entity_cs_ceaf',
        #'cs_b_cubed',
        'b_cubed_plus',
        'typed_mention_ceaf',
        'mention_ceaf_plus',
        'typed_mention_ceaf_plus',
    },
    CORNOLTI_WWW13_MEASURES: [
        'strong_linked_mention_match',
        'strong_link_match',
        'entity_match',
        ],
    HACHEY_ACL14_MEASURES: [
        'strong_mention_match',  # full ner
        'strong_linked_mention_match',
        'strong_link_match',
        'entity_match',
        ],
    LUO_MEASURES: [
        'muc',
        'b_cubed',
        'mention_ceaf',
        'entity_ceaf',
        ],
    #CAI_STRUBE_MEASURES: [
    #    'cs_b_cubed',
    #    'entity_cs_ceaf',
    #    'mention_cs_ceaf',
    #],
    TAC09_MEASURES: [
        'strong_link_match',  # recall equivalent to kb accuracy
        'strong_nil_match',  # recall equivalent to nil accuracy
        'strong_all_match',  # equivalent to overall accuracy
        ],
    TAC11_MEASURES: [
        TAC09_MEASURES,
        'b_cubed', # standard b-cubed
        'b_cubed_plus', # also requires correct resolution
        ],
    TAC14_MEASURES: [
        TAC11_MEASURES,
        # Assess mention recognition in TAC 2014 end-to-end task
        'strong_mention_match', # span must match
        'strong_typed_mention_match', # span and type must match
        # Assess recognition and disambiguation in TAC 2014 end-to-end task
        'strong_typed_all_match',  # span, type and resolution/nil must match
        # Assess recognition and clustering in TAC 2014 end-to-end task
        'mention_ceaf', # prf based on cluster alignment
        'typed_mention_ceaf',  # same requiring type match
    ],
    TMP_MEASURES: [
        'mention_ceaf',
        'entity_ceaf',
        'pairwise',
        ],
}

DEFAULT_MEASURE_SET = ALL_MEASURES
DEFAULT_MEASURE = 'strong_all_match'


def _expand(measures):
    if isinstance(measures, str):
        if measures in MEASURE_SETS:
            measures = MEASURE_SETS[measures]
        else:
            return [measures]
    if isinstance(measures, Measure):
        return [measures]
    if len(measures) == 1:
        return _expand(measures[0])
    return [m for group in measures for m in _expand(group)]


def parse_measures(in_measures, incl_clustering=True, allow_unknown=False):
    """Expand measure lists and validate, but do not map strings to Measures

    """
    # flatten nested sequences and expand group names
    measures = _expand(in_measures)
    # remove duplicates while maintaining order
    seen = set()
    measures = [seen.add(m) or m
                for m in measures if m not in seen]

    # TODO: make sure resolve to valid measures
    not_found = set(measures) - keys(MEASURES)
    invalid = []
    for m in not_found:
        try:
            get_measure(m)
        except Exception:
            invalid.append(m)
    if invalid and not allow_unknown:
        raise ValueError('Could not resolve measures: '
                         '{}'.format(sorted(not_found)))

    if not incl_clustering:
        measures = [m for m in measures
                    if not get_measure(m).is_clustering]
    # TODO: remove clustering metrics given flag
    # raise error if empty
    if not measures:
        msg = 'Could not resolve {!r} to any measures.'.format(in_measures)
        if not incl_clustering:
            msg += ' Clustering measures have been excluded.'
        raise ValueError(msg)
    return measures


def get_measure(name, weighting=None):
    if isinstance(name, Measure):
        measure = name
    elif name.count(':') == 2:
        measure = Measure.from_string(name)
    else:
        measure = MEASURES[name]
    return measure.with_weighting(weighting)


def get_measure_choices():
    return sorted(MEASURE_SETS.keys()) + sorted(MEASURES.keys())


MEASURE_HELP = ('Which measures to use: specify a name (or group name) from '
                'the list-measures command. This flag may be repeated.')


def _wrap(text):
    return '\n'.join(textwrap.wrap(text))


class ListMeasures(object):
    """List measures schemes available for evaluation"""

    def __init__(self, measures=None):
        self.measures = measures

    def __call__(self):
        from .coref_metrics import COREF_METRICS
        measures = parse_measures(self.measures or get_measure_choices())
        header = ['Name', 'Aggregate', 'Filter', 'Key Fields', 'In groups']
        rows = [header]

        set_membership = defaultdict(list)
        for set_name, measure_set in sorted(MEASURE_SETS.items()):
            for name in parse_measures(measure_set):
                set_membership[name].append(set_name)

        for name in sorted(measures):
            measure = get_measure(name)
            rows.append((name, measure.agg, str(measure.filter),
                         '+'.join(measure.key),
                         ', '.join(set_membership[name])))

        col_widths = [max(len(row[i]) for row in rows)
                      for i in range(len(header))]
        rows.insert(1, ['=' * w for w in col_widths])
        fmt = '\t'.join('{:%ds}' % w for w in col_widths[:-1]) + '\t{}'
        ret = _wrap('The following lists possible values for --measure (-m) '
                    'in evaluate, confidence and significance. The name from '
                    'each row or the name of a group may be used. ') + '\n\n'
        ret = '\n'.join(textwrap.wrap(ret)) + '\n\n'
        ret += '\n'.join(fmt.format(*row) for row in rows)
        ret += '\n\nDefault evaluation group: {}'.format(DEFAULT_MEASURE_SET)
        ret += '\n\n'
        ret += _wrap('In all measures, a set of tuples corresponding to Key '
                     'Fields is produced from annotations matching Filter. '
                     'Aggregation with `sets\' compares gold and predicted '
                     'tuple sets directly; coreference aggregates compare '
                     'tuples clustered by their assigned entity ID.')
        ret += '\n\n'
        ret += ('A measure may be specified explicitly. Thus:\n'
                '  {}\nmay be entered as\n  {}'
                ''.format(DEFAULT_MEASURE, get_measure(DEFAULT_MEASURE)))
        ret += '\n\n'
        ret += ('Available aggregates are:\n'
                '- non-clustering: %s\n'
                '- clustering: %s' % (', '.join(sorted(Measure.NON_CLUSTERING_AGG)),
                                      ', '.join(sorted(COREF_METRICS))))
        ret += '\n\n'
        ret += _wrap('Available filter and key fields: %s.' %
                     ', '.join(sorted(Annotation.list_fields())))
        ret += '\n\n'
        ret += _wrap('More fields can be stored dynamically by entering a candidate\'s type as a JSON key-value mapping.')
        return ret

    @classmethod
    def add_arguments(cls, p):
        p.add_argument('-m', '--measure', dest='measures', action='append',
                       metavar='NAME', help=MEASURE_HELP)
        p.set_defaults(cls=cls)
        return p


class TypeWeighting:
    def __init__(self, path):
        self.values = {}
        with open(path) as f:
            for l in f:
                gold, sys, weight = l.split('\t')
                weight = float(weight)
                self.values[gold, sys] = max(weight,
                                             self.values.get((gold, sys),
                                                             weight))

    def __call__(self, gold_type, sys_type):
        return self.values.get((gold_type, sys_type),
                               int(gold_type == sys_type))


def load_weighting(type_weights=None):
    """type_weights is a path
    """
    weighting = {}
    if type_weights is not None:
        weighting['type'] = TypeWeighting(type_weights)
    return weighting or None



import datasets
from typing import List, Dict
from . import init_schema


def from_spacy(docs, span_key=None, doc_id_key=None):
    """
    Converts a list of NER-tagged spaCy documents into a Hugging Face Datasets dataset with the BigBIO schema.

    Args:
    - docs: a list of spaCy documents.
    - span_key: a string specifying the span attribute to use for entity spans. If None, the default 'ents' attribute is used.
    - doc_id_key: a string specifying the key for the document ID to use. If None, document ID is assigned based on position in the input list.

    Returns:
    - a Hugging Face Datasets dataset.
    """
    id_range = list(range(0, len(docs)))

    ds_dict = {
        "id": [],
        "document_id": [],
        "passages": [],
        "entities": [],
    }

    for id, doc in zip(id_range, docs):
        if doc_id_key:
            document_id = doc.user_data[doc_id_key]
        else:
            document_id = id
        ds_dict["id"].append(id)
        ds_dict["document_id"].append(document_id)
        passages = []
        for sid, sent in enumerate(doc.sents):
            passages.append(
                {"id": sid, "type": "sentence", "text": [sent.text], "offsets": [[sent.start_char, sent.end_char]]}
            )
        ds_dict["passages"].append(passages)
        entities = []
        if span_key:
            spans = doc.spans[span_key]
        else:
            spans = doc.ents
        for span_id, span in enumerate(sorted(spans, key=lambda s: s.start_char)):
            entities.append(
                {
                    "id": span_id,
                    "type": span.label_,
                    "text": [span.text],
                    "offsets": [[span.start_char, span.end_char]],
                    "normalized": [],
                }
            )
        ds_dict["entities"].append(entities)

    n_docs = len(ds_dict["document_id"])
    ds_dict["coreferences"] = [[]] * n_docs
    ds_dict["relations"] = [[]] * n_docs
    ds_dict["events"] = [[]] * n_docs
    return datasets.Dataset.from_dict(init_schema(ds_dict))


def from_spans(
    entities: List[List[Dict]], sentences: List[str], document_ids: List[str] = None, sentence_offsets: List[int] = None
):
    """
    Converts a list of spans into a Hugging Face Datasets dataset with the BigBIO schema.

    Args:
        - entities: a list of lists of entities. Each entity is a dictionary with the following keys:
            - char_start_index: the start character index of the entity span.
            - char_end_index: the end character index of the entity span.
            - label: the entity type.
            - span: the entity text span.
        - sentences: a list of sentences.
        - document_ids: a list of document IDs. If None, document IDs are assigned consecutively.
        - sentence_offsets: a list of sentence offsets. If None, sentence offsets are assigned based on sentence length.
    """
    if not document_ids:
        document_ids = [i for i in range(0, len(sentences))]

    # Convert to str
    document_ids = [str(i) for i in document_ids]

    if not sentence_offsets:
        sentence_offsets = []
        prev_doc_id = None
        for doc_id, sent in zip(document_ids, sentences):
            if prev_doc_id is None or prev_doc_id != doc_id:
                offset = 0
            sentence_offsets.append(offset)
            offset += len(sent) + 1
            prev_doc_id = doc_id

    ents = {d: [] for d in document_ids}
    passages = {d: [] for d in document_ids}

    for sentence_entities, sentence, doc_id, sentence_start in zip(entities, sentences, document_ids, sentence_offsets):
        for entity in sentence_entities:
            off0 = entity["char_start_index"] + sentence_start
            off1 = entity["char_end_index"] + sentence_start
            ent = {
                "id": f"{doc_id}_{len(ents[doc_id])}",
                "type": entity["label"],
                "text": [entity["span"]],
                "offsets": [[off0, off1]],
                "normalized": [],
            }
            ents[doc_id].append(ent)
        passage = {
            "text": [sentence],
            "type": "sentence",
            "offsets": [[sentence_start, sentence_start + len(sentence)]],
        }
        passages[doc_id].append(passage)

    out = {"id": [], "document_id": [], "passages": [], "entities": []}

    for doc_id, doc_passages in passages.items():
        out["id"].append(doc_id)
        out["document_id"].append(doc_id)

        for i, passage in enumerate(doc_passages):
            passage["id"] = f"{doc_id}_{i}"
        out["passages"].append(doc_passages)

        doc_ents = ents[doc_id]
        out["entities"].append(doc_ents)

    n_docs = len(out["document_id"])
    out["coreferences"] = [[]] * n_docs
    out["relations"] = [[]] * n_docs
    out["events"] = [[]] * n_docs
    return datasets.Dataset.from_dict(init_schema(out))



from itertools import groupby


class Deduplicator:
    def _deduplicate(self, document):
        result = []
        grp_key = lambda e: (e["offsets"], e["type"], e["text"])
        # sorted_ents = sorted(document["entities"], key=grp_key)
        for _, grp in groupby(document["entities"], grp_key):
            for i, g in enumerate(grp):
                g["normalized"] = g["normalized"][i:]
                result.append(g)
        return {"entities": result}

    def transform_batch(self, ds):
        return ds.map(lambda e: self._deduplicate(e), load_from_cache_file=False)



from pathlib import Path
from omegaconf import OmegaConf


def load_config(file_name):
    """
    Loads and returns the configuration stored in a YAML file using OmegaConf.

    Args:
    - file_name (str): The path to the YAML file containing the configuration.

    Returns:
    - An OmegaConf object representing the YAML file's contents.

    Note:
        If the configuration file has a `base_config` attribute, this function will attempt to load and merge the base
        configuration file with the current configuration file. The base configuration file should be a YAML file as well,
        and its path is relative to the current configuration file's directory.

    """
    path = Path(file_name)
    conf = OmegaConf.load(path)
    if base_config_path := conf.get("base_config", None):
        base_conf = load_config(path.parent / base_config_path)
        conf = OmegaConf.merge(base_conf, conf)
    return conf



from typing import Union
from pathlib import Path
from tqdm import tqdm

from datasets import DatasetDict

from xmen.reranking import Reranker, multiple_choice_util, ranking_util
from xmen.data import IndexedDataset, IndexedDatasetDict

from transformers import (
    TrainingArguments,
    Trainer,
    AutoModelForMultipleChoice,
)


def index_of(val, in_list):
    """
    Returns the index of the first occurrence of the specified value in the given list.
    If the value is not found, returns 0.

    Args:
    - val: the value to search for
    - in_list: the list to search in

    Returns:
    - The index of the first occurrence of the specified value in the given list.
      If the value is not found, returns 0.
    """
    try:
        return in_list.index(val) + 1
    except ValueError:
        return 0


def preprocess_el_dataset(examples, tokenizer, context_length: int, n_candidates: int, encode_positions: bool = False):
    """
    Preprocesses a dataset of examples for entity linking, converting them to tokenized input
    suitable for use with a neural network model. Each example consists of a mention and a list of
    candidate entities, represented as their synonyms.

    Args:
    - examples: a dictionary containing the "mention", "synonyms", "context_left", and "context_right"
               fields for each example.
    - tokenizer: the tokenizer to use for tokenizing the input
    - context_length: the maximum length of context to include on either side of the mention, in tokens
    - n_candidates: the number of candidates to include for each mention
    - encode_positions: if True, includes position encoding in the output tokens

    Returns:
    - A dictionary containing the tokenized examples, split into inputs for each mention and its
    corresponding candidate entities. The keys are the names of the inputs ("input_ids",
    "attention_mask", "token_type_ids"), and the values are lists of lists, where each inner list
    corresponds to the inputs for a single mention and its candidate entities.
    """
    actual_n_candidates = n_candidates + 1  # if candidate not included
    mentions = [
        [
            f"{l_ctx[-context_length:] if context_length > 0 else ''} [START] {m} [END] {r_ctx[:context_length] if context_length > 0 else ''}"
        ]
        * actual_n_candidates
        for l_ctx, m, r_ctx in zip(examples["context_left"], examples["mention"], examples["context_right"])
    ]
    mentions = sum(mentions, [])
    candidates = []
    for e in examples["synonyms"]:
        if len(e) < n_candidates:
            e.extend([""] * (n_candidates - len(e)))
        e.insert(0, ["[UNK]"])

    candidates = [
        (f"[POSSTART] {i} [POSEND] " if encode_positions else "") + (" [SEP] ".join(s))
        for e in examples["synonyms"]
        for i, s in enumerate(e)
    ]

    tokenized_examples = tokenizer(mentions, candidates, truncation=True, max_length=512)

    return {
        k: [v[i : i + actual_n_candidates] for i in range(0, len(v), actual_n_candidates)]
        for k, v in tokenized_examples.items()
    }


def create_multiple_choice_dataset(
    candidate_ds,
    ground_truth,
    kb,
    tokenizer,
    context_length: int,
    k: int,
    expand_abbreviations: bool,
):
    """
    Creates a multiple choice dataset suitable for training a neural network model for entity linking.
    The dataset consists of a list of mentions and their corresponding candidate entities, where one of
    the candidates is the correct entity for the mention. The dataset is tokenized and indexed, ready for
    use with a neural network.

    Args:
    - candidate_ds: the dataset of candidate entities for each mention
    - ground_truth: the correct entity for each mention
    - kb: the knowledge base to use for entity linking
    - tokenizer: the tokenizer to use for tokenizing the input
    - context_length: the maximum length of context to include on either side of the mention, in tokens
    - k: the number of candidates to include for each mention
    - expand_abbreviations: if True, expands abbreviations in the dataset using the knowledge base

    Returns:
    - An IndexedDataset object containing the tokenized and indexed multiple choice dataset.
    """
    flat_candidate_ds, doc_index = ranking_util.get_flat_candidate_ds(
        candidate_ds, ground_truth, expand_abbreviations, kb
    )

    # get index of the correct answer or 0 if not in the dataset
    flat_candidate_ds = flat_candidate_ds.map(
        lambda e: {"label": index_of(e["label"][0], e["candidates"])},
        load_from_cache_file=False,
    )

    ds_tokenized = flat_candidate_ds.map(
        lambda e: preprocess_el_dataset(e, tokenizer, context_length, k),
        batched=True,
        remove_columns=[c for c in flat_candidate_ds.column_names if c != "label"],
        load_from_cache_file=False,
    )
    return IndexedDataset(ds_tokenized, doc_index)


class ListWiseReranker(Reranker):
    def __init__(self):
        pass

    @staticmethod
    def prepare_data(
        candidates,
        ground_truth,
        kb,
        tokenizer,
        context_length: int,
        k: int,
        expand_abbreviations: bool,
        **kwargs,
    ):
        """
        Given a list of candidates, their corresponding ground truth labels, a knowledge base (kb), a tokenizer, and other optional arguments, creates a dataset that can be used to train a listwise reranker. Returns an IndexedDatasetDict containing train, validation, and/or test datasets depending on the input types.

        Args:
        - candidates: Union[DatasetDict, List[Dict[str, Union[str, List[str]]]]] - A dataset or list of dictionaries containing the candidates, their labels, and optionally other metadata.
        - ground_truth: Optional[DatasetDict] - A dataset containing the ground truth labels for the candidates. Only necessary if candidates is a DatasetDict.
        - kb: Union[Path, str] - A path to a file or directory containing the knowledge base for the task.
        - tokenizer: PreTrainedTokenizerBase - A tokenizer object to use for tokenizing the candidates and ground truth labels.
        - context_length: int - The maximum number of tokens to include in the context of each candidate during tokenization.
        - k: int - The number of candidates to consider in each listwise comparison.
        - expand_abbreviations: bool - Whether to expand abbreviations in the candidates and ground truth labels during tokenization.
        - **kwargs: Additional optional keyword arguments to pass to create_multiple_choice_dataset().

        Returns:
        - res: IndexedDatasetDict - An IndexedDatasetDict containing the train, validation, and/or test datasets.

        Raises:
        - AssertionError if `candidates` is not a DatasetDict
        """
        print("Context length:", context_length)
        print("n candidates:", k)
        if type(candidates) == DatasetDict:
            assert type(ground_truth) == DatasetDict
            res = IndexedDatasetDict()
            for split, cand in candidates.items():
                gt = ground_truth[split]
                ds = create_multiple_choice_dataset(
                    cand,
                    gt,
                    kb,
                    tokenizer,
                    context_length,
                    k,
                    expand_abbreviations,
                )
                res[split] = ds
            return res
        else:
            return create_multiple_choice_dataset(
                candidates,
                ground_truth,
                kb,
                tokenizer,
                context_length,
                k,
                expand_abbreviations,
            )

    @staticmethod
    def get_trainer(
        train_dataset,
        val_dataset,
        tokenizer,
        output_dir: Union[str, Path],
        **training_args,
    ):
        """
        Given train and validation datasets, a tokenizer, an output directory, and other optional training arguments, creates a Trainer object that can be used to train a listwise reranker. Returns the Trainer object.

        Args:
        - train_dataset: Dataset - A dataset containing the training examples.
        - val_dataset: Dataset - A dataset containing the validation examples.
        - tokenizer: PreTrainedTokenizerBase - A tokenizer object to use for tokenizing the examples.
        - output_dir: Union[Path, str] - The path to the output directory where the trained model will be saved.
        - **training_args: Additional optional keyword arguments to pass to the Trainer object.

        Returns:
        - trainer: Trainer - A Trainer object that can be used to train a listwise reranker.
        """
        model = AutoModelForMultipleChoice.from_pretrained(training_args["model_name"])

        training_args = TrainingArguments(
            output_dir=output_dir,
            overwrite_output_dir=True,
            evaluation_strategy="epoch",
            # save_strategy="no",
            save_strategy="epoch",
            gradient_checkpointing=True,
            load_best_model_at_end=True,
            report_to="wandb",
            save_total_limit=2,
            run_name=training_args.get("run_name", None),
            metric_for_best_model="f1",
            **{k: v for k, v in training_args.items() if k != "model_name"},
        )

        trainer = Trainer(
            model=model,
            args=training_args,
            train_dataset=train_dataset,
            eval_dataset=val_dataset,
            tokenizer=tokenizer,
            data_collator=multiple_choice_util.DataCollatorForMultipleChoice(tokenizer=tokenizer),
            compute_metrics=multiple_choice_util.compute_entity_linking_metrics,
        )

        return trainer

    def rerank_batch(self):
        pass



#!/usr/bin/env python
import numpy as np
import argparse
import torch
import torch.optim as optim
from torch.cuda.amp import autocast 
from torch.cuda.amp import GradScaler
from pytorch_metric_learning import samplers
import logging
import time
import pdb
import os
import json
import random
from tqdm import tqdm

import sys
sys.path.append("../") 

import wandb
wandb.init(project="sapbert")

from src.data_loader import (
    DictionaryDataset,
    QueryDataset,
    QueryDataset_pretraining,
    MetricLearningDataset,
    MetricLearningDataset_pairwise,
)
from xmen.linkers.model_wrapper import (
    Model_Wrapper
)
from src.metric_learning import (
    Sap_Metric_Learning,
)

LOGGER = logging.getLogger()

def parse_args():
    """
    Parse input arguments
    """
    parser = argparse.ArgumentParser(description='sapbert train')

    # Required
    parser.add_argument('--model_dir', 
                        help='Directory for pretrained model')
    parser.add_argument('--train_dir', type=str, required=True,
                    help='training set directory')
    parser.add_argument('--output_dir', type=str, required=True,
                        help='Directory for output')
    
    # Tokenizer settings
    parser.add_argument('--max_length', default=25, type=int)

    # Train config
    parser.add_argument('--use_cuda',  action="store_true")
    parser.add_argument('--learning_rate',
                        help='learning rate',
                        default=0.0001, type=float)
    parser.add_argument('--weight_decay',
                        help='weight decay',
                        default=0.01, type=float)
    parser.add_argument('--train_batch_size',
                        help='train batch size',
                        default=240, type=int)
    parser.add_argument('--epoch',
                        help='epoch to train',
                        default=3, type=int)
    parser.add_argument('--save_checkpoint_all', action="store_true")
    parser.add_argument('--checkpoint_step', type=int, default=10000000)
    parser.add_argument('--amp', action="store_true", 
            help="automatic mixed precision training")
    parser.add_argument('--parallel', action="store_true") 
    #parser.add_argument('--cased', action="store_true") 
    parser.add_argument('--pairwise', action="store_true",
            help="if loading pairwise formatted datasets") 
    parser.add_argument('--random_seed',
                        help='epoch to train',
                        default=1996, type=int)
    parser.add_argument('--loss',
                        help="{ms_loss|cosine_loss|circle_loss|triplet_loss}}",
                        default="ms_loss")
    parser.add_argument('--use_miner', action="store_true") 
    parser.add_argument('--miner_margin', default=0.2, type=float) 
    parser.add_argument('--type_of_triplets', default="all", type=str) 
    parser.add_argument('--agg_mode', default="cls", type=str, help="{cls|mean|mean_all_tok}") 

    args = parser.parse_args()
    return args

def init_logging():
    LOGGER.setLevel(logging.INFO)
    fmt = logging.Formatter('%(asctime)s: [ %(message)s ]',
                            '%m/%d/%Y %I:%M:%S %p')
    console = logging.StreamHandler()
    console.setFormatter(fmt)
    LOGGER.addHandler(console)

def init_seed(seed=None):
    if seed is None:
        seed = int(round(time.time() * 1000)) % 10000

    LOGGER.info("Using seed={}, pid={}".format(seed, os.getpid()))
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    random.seed(seed)
    torch.backends.cudnn.deterministic = True
    
def load_dictionary(dictionary_path):
    """
    load dictionary
    
    Parameters
    ----------
    dictionary_path : str
        a path of dictionary
    """
    dictionary = DictionaryDataset(
            dictionary_path = dictionary_path
    )
    
    return dictionary.data
    
def load_queries(data_dir, filter_composite, filter_duplicate):
    """
    load query data
    
    Parameters
    ----------
    is_train : bool
        train or dev
    filter_composite : bool
        filter composite mentions
    filter_duplicate : bool
        filter duplicate queries
    """
    dataset = QueryDataset(
        data_dir=data_dir,
        filter_composite=filter_composite,
        filter_duplicate=filter_duplicate
    )
    
    return dataset.data

def load_queries_pretraining(data_dir, filter_duplicate):
    """
    load query data
    
    Parameters
    ----------
    is_train : bool
        train or dev
    filter_composite : bool
        filter composite mentions
    filter_duplicate : bool
        filter duplicate queries
    """
    dataset = QueryDataset_pretraining(
        data_dir=data_dir,
        filter_duplicate=filter_duplicate
    )
    
    return dataset.data

def train(args, data_loader, model, scaler=None, model_wrapper=None, step_global=0):
    LOGGER.info("train!")
    
    train_loss = 0
    train_steps = 0
    model.cuda()
    model.train()
    for i, data in tqdm(enumerate(data_loader), total=len(data_loader)):
        model.optimizer.zero_grad()

        batch_x1, batch_x2, batch_y = data
        batch_x_cuda1, batch_x_cuda2 = {},{}
        for k,v in batch_x1.items():
            batch_x_cuda1[k] = v.cuda()
        for k,v in batch_x2.items():
            batch_x_cuda2[k] = v.cuda()

        batch_y_cuda = batch_y.cuda()
    
        if args.amp:
            with autocast():
                loss = model(batch_x_cuda1, batch_x_cuda2, batch_y_cuda)  
        else:
            loss = model(batch_x_cuda1, batch_x_cuda2, batch_y_cuda)  
        if args.amp:
            scaler.scale(loss).backward()
            scaler.step(model.optimizer)
            scaler.update()
        else:
            loss.backward()
            model.optimizer.step()

        train_loss += loss.item()
        wandb.log({"Loss": loss.item()})
        train_steps += 1
        step_global += 1
        #if (i+1) % 10 == 0:
        #LOGGER.info ("epoch: {} loss: {:.3f}".format(i+1,train_loss / (train_steps+1e-9)))
        #LOGGER.info ("epoch: {} loss: {:.3f}".format(i+1, loss.item()))

        # save model every K iterations
        if step_global % args.checkpoint_step == 0:
            checkpoint_dir = os.path.join(args.output_dir, "checkpoint_iter_{}".format(str(step_global)))
            if not os.path.exists(checkpoint_dir):
                os.makedirs(checkpoint_dir)
            model_wrapper.save_model(checkpoint_dir)
    train_loss /= (train_steps + 1e-9)
    return train_loss, step_global
    
def main(args):
    init_logging()
    #init_seed(args.seed)
    print(args)

    torch.manual_seed(args.random_seed)
    
    # prepare for output
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)
        

    # load BERT tokenizer, dense_encoder
    model_wrapper = Model_Wrapper()
    encoder, tokenizer = model_wrapper.load_bert(
        path=args.model_dir,
        max_length=args.max_length,
        use_cuda=args.use_cuda,
        #lowercase=not args.cased
    )
    
    # load SAP model
    model = Sap_Metric_Learning(
            encoder = encoder,
            learning_rate=args.learning_rate, 
            weight_decay=args.weight_decay,
            use_cuda=args.use_cuda,
            pairwise=args.pairwise,
            loss=args.loss,
            use_miner=args.use_miner,
            miner_margin=args.miner_margin,
            type_of_triplets=args.type_of_triplets,
            agg_mode=args.agg_mode,
    )

    if args.parallel:
        model.encoder = torch.nn.DataParallel(model.encoder)
        LOGGER.info("using nn.DataParallel")
    
    def collate_fn_batch_encoding(batch):
        query1, query2, query_id = zip(*batch)
        query_encodings1 = tokenizer.batch_encode_plus(
                list(query1), 
                max_length=args.max_length, 
                padding="max_length", 
                truncation=True,
                add_special_tokens=True,
                return_tensors="pt")
        query_encodings2 = tokenizer.batch_encode_plus(
                list(query2), 
                max_length=args.max_length, 
                padding="max_length", 
                truncation=True,
                add_special_tokens=True,
                return_tensors="pt")
        #query_encodings_cuda = {}
        #for k,v in query_encodings.items():
        #    query_encodings_cuda[k] = v.cuda()
        query_ids = torch.tensor(list(query_id))
        return  query_encodings1, query_encodings2, query_ids

    if args.pairwise:
        train_set = MetricLearningDataset_pairwise(
                path=args.train_dir,
                tokenizer = tokenizer
        )
        train_loader = torch.utils.data.DataLoader(
            train_set,
            batch_size=args.train_batch_size,
            shuffle=True,
            num_workers=16,
            collate_fn=collate_fn_batch_encoding
        )
    else:
        train_set = MetricLearningDataset(
            path=args.train_dir,
            tokenizer = tokenizer
        )
        # using a sampler
        train_loader = torch.utils.data.DataLoader(
            train_set,
            batch_size=args.train_batch_size,
            #shuffle=True,
            sampler=samplers.MPerClassSampler(train_set.query_ids,\
                2, length_before_new_iter=100000),
            num_workers=16, 
            )
    # mixed precision training 
    if args.amp:
        scaler = GradScaler()
    else:
        scaler = None

    start = time.time()
    step_global = 0
    for epoch in range(1,args.epoch+1):
        # embed dense representations for query and dictionary for train
        # Important! This is iterative process because dense represenation changes as model is trained.
        LOGGER.info("Epoch {}/{}".format(epoch,args.epoch))

        # train
        train_loss, step_global = train(args, data_loader=train_loader, model=model, scaler=scaler, model_wrapper=model_wrapper, step_global=step_global)
        LOGGER.info('loss/train_per_epoch={}/{}'.format(train_loss,epoch))
        
        # save model every epoch
        if args.save_checkpoint_all:
            checkpoint_dir = os.path.join(args.output_dir, "checkpoint_{}".format(epoch))
            if not os.path.exists(checkpoint_dir):
                os.makedirs(checkpoint_dir)
            model_wrapper.save_model(checkpoint_dir)
        
        # save model last epoch
        if epoch == args.epoch:
            model_wrapper.save_model(args.output_dir)
            
    end = time.time()
    training_time = end-start
    training_hour = int(training_time/60/60)
    training_minute = int(training_time/60 % 60)
    training_second = int(training_time % 60)
    LOGGER.info("Training Time!{} hours {} minutes {} seconds".format(training_hour, training_minute, training_second))
    
if __name__ == '__main__':
    args = parse_args()
    main(args)



import typer
import os
import logging
from pathlib import Path

from .dict import *
from .index import *
from .utils import *
from ..log import logger


app = typer.Typer()


@app.command("dict")
def assemble_kb(
    cfg_path: str = typer.Argument(..., help=cfg_path_help, show_default=False),
    code: str = typer.Option(default=None, help=code_help, show_default=True),
    output: str = typer.Option(default=None, help=output_help, show_default=True),
    key: str = typer.Option(default=None, help=key_help, show_default=True),
    log_file: str = typer.Option(default=None, help=log_file_help, show_default=True),
    overwrite: bool = typer.Option(default=False, help=overwrite_help, show_default=True),
    custom_name: str = typer.Option(default=None, help=custom_name_help, show_default=True),
):
    """
    Builds a .jsonl dictionary with the given configuration for UMLS or custom knowledge bases.

    Args:
    - cfg_path (str): Path to the configuration file.
    - code (Optional[str]): Path to the custom parsing script.
    - output (Optional[str]): Path to the output directory.
    - key (Optional[str]): Name of the subconfig.
    - log_file (Optional[str]): Path to the log file.
    - overwrite (bool): Flag indicating whether to overwrite the existing file in the output directory.
    - custom_name (Optional[str]): Name of the custom knowledge base.

    Raises:
    - typer.Exit: If the provided config file does not contain `umls` nor `custom` keys nested within `dict`,
            and no key is provided with --key.
    """
    if log_file is not None:
        add_file_logger(log_file)
    logger.debug("\n \n NEW DICT RUN")
    cfg = check_and_load_config(cfg_path)

    # make sure there is umls or custom or subconfig key in dict
    if "umls" not in cfg.dict and "custom" not in cfg.dict and not key:
        logger.error(
            "The provided config file does not contain `umls` nor `custom` keys nested within `dict`."
            "If you are attempting to use one of multiple subconfigs, provided it's key name with --key."
        )
        raise typer.Exit()

    # if a subconfig key is given, prune the rest of them out
    if key:
        cfg.dict = cfg.dict[key]
        cfg.name = key

    # custom_name prevails over the key name of a subconfig
    name = custom_name if custom_name is not None else cfg.name
    dict_dir = get_work_dir(cfg) / f"{name}.jsonl" if output is None else Path(output) / f"{name}.jsonl"

    if can_write(dict_dir, overwrite):
        # handle umls kbs
        if "umls" in cfg.dict:
            logger.info(f"Attempting to build {os.path.basename(cfg_path)} in {dict_dir}. This can take some minutes.")
            assemble(cfg, output=dict_dir)
            logger.info(f"Success! {os.path.basename(cfg_path)} assembled in {dict_dir}")
        # handle non-umls without (or wrong) custom code provided
        elif code is None:
            logger.error(
                f"No custom code provided for non-UMLS kb {name}. To use custom code to parse {name}, use flag"
                f" `--code path-to-script/script.py`. If {name} uses UMLS, the problem may be in the provided "
                f".yaml file: the expected file should have key `umls` nested within key `dict`"
            )
        # handle non-umls kbs with correct custom parsing script
        elif code is not None or "custom" in cfg.dict:
            logger.info(
                f"Attempting to build {os.path.basename(cfg_path)} from custom script in {code} to {dict_dir}. This "
                "can take some minutes."
            )
            assemble(cfg, output=dict_dir, custom_path=code)
            logger.info(f"Success! {os.path.basename(cfg_path)} assembled in {dict_dir}")
    else:
        logger.info("Exiting app without building knowledge base dictionary.")
        raise typer.Exit()

    pass


@app.command("index")
def build_ngram_sapbert(
    cfg_path: str = typer.Argument(..., help=cfg_path_help, show_default=False),
    gpu_id: int = typer.Option(default=0, help=gpu_id_help, show_default=True),
    output: str = typer.Option(default=None, help=output_help, show_default=True),
    log_file: str = typer.Option(default=None, help=log_file_help, show_default=True),
    dict: str = typer.Option(default=None, help=dict_dir_help, show_default=True),
    overwrite: bool = typer.Option(default=False, help=overwrite_help, show_default=True),
    sapbert: bool = typer.Option(default=False, help=sapbert_help, show_default=True),
    ngram: bool = typer.Option(default=False, help=ngram_help, show_default=True),
    all: bool = typer.Option(default=False, help=all_help, show_default=True),
    batch_size: int = typer.Option(default=DEFAULT_BATCH_SIZE, help=batch_help, show_default=True),
    index_buffer_size: int = typer.Option(
        default=DEFAULT_INDEX_BUFFER_SIZE, help=index_buffer_size_help, show_default=True
    ),
    save_ram: bool = typer.Option(default=False, help=save_ram_help, show_default=True),
):
    """
    Builds N-Gram and SAPBert indices from a .jsonl dict with the given configuration.

    Args:
    - cfg_path (str): Path to the configuration file.
    - gpu_id (int): Index of the GPU to be used for building the SAPBert indices.
    - output (str): Path to the directory where the indices will be stored. If None, the work directory will be used.
    - log_file (str): Path to the file where logs will be written. If None, logs will be printed to console.
    - dict (str): Path to the .jsonl dictionary. If None, the work directory and configuration file name will be used to search for the dictionary.
    - overwrite (bool): Whether to overwrite existing indices or not.
    - sapbert (bool): Whether to build SAPBert indices or not.
    - ngram (bool): Whether to build N-Gram indices or not.
    - all (bool): Whether to build both SAPBert and N-Gram indices or not.

    Raises:
    - typer.Exit: If neither sapbert nor ngram nor all is set to True.
    """
    if log_file is not None:
        add_file_logger(log_file)
    logger.debug("\n \n NEW INDEX RUN")
    cfg = check_and_load_config(cfg_path)
    work_dir = get_work_dir(cfg)
    output = work_dir if output is None else Path(output)
    dict = work_dir / f"{cfg.name}.jsonl" if dict is None else dict

    if not sapbert and not ngram and not all:
        logger.warn("Please indicate which indicess to build with --sapbert or --ngram or --all.")
        raise typer.Exit()

    if os.path.exists(dict):
        logger.info(f"Dictionary found in {dict}. Attempting to build indices.")

        is_ngram_selected = ngram or all
        write_path = output / "index" / "ngrams"
        if is_ngram_selected and can_write(write_path, overwrite):
            logger.info("Building N-Gram index.")
            build_ngram(cfg, output, dict)
        else:
            logger.info("Skipping N-Gram index.")

        is_sapbert_selected = sapbert or all
        write_path = output / "index" / "sapbert"
        if is_sapbert_selected and can_write(write_path, overwrite):
            # check for GPUs
            if torch.cuda.is_available():
                logger.info(f"CUDA is available. Running on GPU with ID {gpu_id}. To select another, use --gpu-id.")
            else:
                gpu_id = -1
                logger.warning(
                    "CUDA is not available on this system. Running on CPU. This can take considerably longer."
                )
            logger.info("Building SapBERT index.")
            build_sapbert(cfg, output, dict, gpu_id, batch_size, index_buffer_size, save_ram)
        else:
            logger.info("Skipping SapBERT index.")

    else:
        logger.info(
            f"Dictionary {dict} does not exist. Build one with command --dict or provide the right path to one with "
            "--dict-dir flag."
        )

    pass


def main():
    app()



#!/usr/bin/env python
"Representation of link standoff annotation and measures over it"
"Adapted from https://github.com/wikilinks/neleval/blob/master/neleval/ for Python >3.10 compatibility"

from collections.abc import Sequence
from collections import defaultdict
import operator
import warnings
import json

from neleval.utils import unicode


try:
    keys = dict.viewkeys
    import itertools
    filter = itertools.ifilter
except Exception:
    # Py3k
    keys = dict.keys


class Annotation(object):
    """A name mention and its predicted candidates

    Parameters
    ----------

    Attributes
    ----------
    TODO
    is_first : bool or unset
        Indicates if this annotation is the first in the document for this
        eid.

    * :
        Other attributes are automatically adopted from the top candidate
    """

    __slots__ = ['docid', 'start', 'end', 'candidates', 'is_first']

    def __init__(self, docid, start, end, candidates=[]):
        self.docid = docid
        self.start = start
        self.end = end
        self.candidates = candidates

    def __unicode__(self):
        return u'{}\t{}\t{}\t{}'.format(
            self.docid,
            self.start,
            self.end,
            u'\t'.join([unicode(c) for c in self.candidates])
            )

    __str__ = __unicode__

    def __repr__(self):
        return 'Annotation({!r}, {!r}, {!r}, {!r})'.format(self.docid, self.start, self.end, self.candidates)

    def __lt__(self, other):
        assert isinstance(other, Annotation)
        return (self.start, -self.end) < (other.start, -other.end)

    def compare_spans(self, other):
        assert self.start <= self.end, 'End is before start: {!r}'.format(self)
        assert other.start <= other.end, 'End is before start: {!r}'.format(self)
        if self.docid != other.docid:
            return 'different documents'
        if self.start > other.end or self.end < other.start:
            return 'non-overlapping'
        elif self.start == other.start and self.end == other.end:
            return 'duplicate'
        elif self.start < other.start and self.end >= other.end:
            return 'nested'
        elif self.start >= other.start and self.end < other.end:
            return 'nested'
        else:
            return 'crossing'

    # Getters
    @property
    def span(self):
        return (self.docid, self.start, self.end)

    @property
    def link(self):
        "Return top candidate"
        if self.candidates:
            return self.candidates[0]

    def __getattr__(self, name):
        # Generally, return attributes from top candidate
        # or None if there are no candidates
        if name.startswith('_'):
            # AttributeError
            super(Annotation, self).__getattr__(name)
        link = self.link
        if link is not None:
            return getattr(link, name)

    # Parsing methods
    @classmethod
    def from_string(cls, s):
        docid, start, end, candidates = None, None, None, []
        cols = s.rstrip('\n\t').split('\t', 3)
        if len(cols) < 3:
            raise SyntaxError('Annotation must have at least 3 columns. Got {!r}'.format(s))
        if len(cols) >= 3:
            docid = cols[0]
            start = int(cols[1])
            end = int(cols[2])
        if len(cols) == 4:
            candidates = sorted(Candidate.from_string(cols[3]), reverse=True)
        return Annotation(docid, start, end, candidates)

    @classmethod
    def list_fields(cls):
        ann = cls.from_string('a\t0\t1\tabc')
        return [f for f in dir(ann) + dir(ann.link)
                if not f.startswith('_')
                and not callable(getattr(ann, f, None))]


class Candidate(object):
    """A candidate link

    Parameters
    ----------
    eid : string
        KB ID or NIL ID (begins "NIL")
    score : numeric
        Confidence
    type : string or dict
        If a string, the ``type`` attribute is set to this string.
        If a dict, all keys are copied to attributes of the candidate

    Attributes
    ----------
    eid : string
    score : numeric
    kbid : string or None
        ``eid`` if it does not begin with "NIL"
    is_nil : bool
        True if ``eid`` begins with "NIL"
    is_linked : bool
        ``not is_nil``

    Examples
    --------
    >>> cand = Candidate.from_string('NIL123')
    >>> (cand.eid, cand.score, cand.kbid, cand.is_nil, cand.is_linked)
    ('NIL123', None, None, True, False)
    >>> cand = Candidate.from_string('E123\t1.5\tPER')
    >>> (cand.eid, cand.score, cand.kbid, cand.is_nil, cand.is_linked)
    ('E123', 1.5, 'E123', False, True)
    >>> cand.type
    'PER'
    >>> cand = Candidate.from_string('E123\t1.5\t{"type":"PER","reftype":"NOM"}')
    >>> cand.type
    'PER'
    >>> cand.reftype
    'NOM'
    """
    __slots__ = ['eid',    # KB ID or NIL ID
                 'score',  # confidence
                 '__dict__',
                 ]

    def __init__(self, eid, score=None, type=None):
        self.eid = eid
        self.score = score
        if hasattr(type, 'items'):
            self.__dict__.update(type)
        else:
            self.type = type

    @property
    def kbid(self):
        "Return link KB ID or None"
        if self.is_linked:
            return self.eid

    @property
    def is_nil(self):
        if self.eid is None:
            return True
        if self.eid.startswith('NIL'):
            return True
        return False

    @property
    def is_linked(self):
        return not self.is_nil

    def __unicode__(self):
        type_ = self.__dict__
        if not type_:
            type_ = ''
        elif len(type_) == 1 and 'type' in type_:
            type_ = type_['type'] or ''
        else:
            type_ = json.dumps(type_)
        return u'{}\t{}\t{}'.format(self.eid,
                                    self.score or '',
                                    type_)

    __str__ = __unicode__

    def __repr__(self):
        return '<{!r}>'.format(self.eid)

    def __lt__(self, other):
        assert isinstance(other, Candidate)
        return self.score < other.score

    # Parsing methods
    @classmethod
    def from_string(cls, s):
        cols = s.rstrip('\t').split('\t')
        if len(cols) == 1:
            # link includes id only
            yield cls(cols[0])
        elif len(cols) == 2:
            # link includes id and score
            yield cls(cols[0], float(cols[1]))
        elif len(cols[3:]) % 3 == 0:
            # >=1 (id, score, type) candidate tuples
            for i in range(0, len(cols), 3):
                id, score, type = cols[i:i+3]
                if type.startswith('{'):
                    type = json.loads(type)
                yield cls(id, float(score), type)
        else:
            # undefined format
            raise SyntaxError('Need id, score and type when >1 candidates')


class Measure(object):
    def __init__(self, key, filter=None, agg='sets', weighting=None):
        """
        key : list of fields for mention comparison
        filter : a function or attribute name to select evaluated annotations
        agg : sets or clustering measure
        weighting : mapping field -> (func(gold_val, pred_val) -> float)
        """
        if not isinstance(key, Sequence):
            raise TypeError('key should be a list or tuple')
        self.key = tuple(key)
        self.filter = filter
        if filter is not None and not callable(filter):
            assert isinstance(filter, str)
            filter = operator.attrgetter(filter)
        self.filter_fn = filter
        if agg.endswith('-micro'):
            self.display_agg = agg
            agg = agg[:-6]
            warnings.warn('`{0}-micro\' aggregate has been renamed to '
                          '`{0}\' and will be removed in a future '
                          'release.'.format(agg),
                          DeprecationWarning)

        self.agg = agg

        if agg != 'sets' and weighting:
            raise NotImplementedError('weighting is only implemented for '
                                      'aggregate="sets"')
        self.weighting = weighting

    def with_weighting(self, weighting):
        return Measure(self.key, self.filter, self.agg, weighting)

    def __str__(self):
        return '{}:{}:{}'.format(getattr(self, 'display_agg', self.agg),
                                 self.filter, '+'.join(self.key))

    @classmethod
    def from_string(cls, s, weighting=None):
        if s.count(':') != 2:
            raise ValueError('Expected 2 colons in {!r}'.format(s))
        a, f, k = s.split(':')
        if f in ('', 'None'):
            f = None
        return cls(k.split('+'), f, a, weighting=weighting)

    def __repr__(self):
        return ('{0.__class__.__name__}('
                '{0.key!r}, {0.filter!r}, {0.agg!r})'.format(self))

    NON_CLUSTERING_AGG = ('sets',) + tuple(
        ['overlap-%s%s' % (p1, p2)
         for p1 in ('sum', 'max')
         for p2 in ('sum', 'max')])

    @property
    def is_clustering(self):
        return self.agg not in self.NON_CLUSTERING_AGG

    def build_index(self, annotations, key_fields=None, multi=False):
        if isinstance(annotations, dict):
            # assume already built
            return annotations
        # TODO: caching

        if self.filter is not None:
            annotations = filter(self.filter_fn, annotations)
        key = self.key if key_fields is None else key_fields
        if multi:
            out = defaultdict(list)
            for ann in annotations:
                out[tuple(getattr(ann, field) for field in key)].append(ann)
            out.default_factory = None
            return out
        else:
            return {tuple(getattr(ann, field) for field in key): ann
                    for ann in annotations}

    def build_clusters(self, annotations):
        if isinstance(annotations, dict):
            # assume already built
            return annotations
        # TODO: caching
        # TODO: can reuse build_index for small efficiency loss

        if self.filter is not None:
            annotations = filter(self.filter_fn, annotations)
        key = self.key
        out = defaultdict(set)
        for ann in annotations:
            out[ann.eid].add(tuple(getattr(ann, field) for field in key))
        out.default_factory = None  # disable defaulting
        return out

    def count_matches(self, system, gold):
        if self.is_clustering:
            raise ValueError('count_matches is inappropriate '
                             'for {}'.format(self.agg))
        if not self.weighting:
            gold_index = self.build_index(gold)
            pred_index = self.build_index(system)
            tp = len(keys(gold_index) & keys(pred_index))

            fn = len(gold_index) - tp
            fp = len(pred_index) - tp
        else:
            key = [f for f in self.key
                   if f not in self.weighting]
            weighting_fields = [f for f in self.key
                                if f in self.weighting]
            gold_index = self.build_index(gold, key_fields=key, multi=True)
            pred_index = self.build_index(system, key_fields=key, multi=True)
            if (
                    any(len(v) > 1 for v in gold_index.values()) or
                    any(len(v) > 1 for v in pred_index.values())
            ):
                raise NotImplementedError('No weighting support where '
                                          'annotations may have duplicate key')
            tp = 0.
            for k, (gold_ann,) in gold_index.items():
                try:
                    pred_ann, = pred_index[k]
                except KeyError:
                    continue
                tp += self.calc_match_weight(gold_ann, pred_ann,
                                             weighting_fields)

            fn = sum(self.calc_match_weight(gold_ann, gold_ann, weighting_fields)
                     for gold_ann, in gold_index.values()) - tp
            fp = sum(self.calc_match_weight(pred_ann, pred_ann, weighting_fields)
                     for pred_ann, in pred_index.values()) - tp
        return tp, fp, fn

    def calc_match_weight(self, gold_ann, pred_ann, weighting_fields):
        w = 1
        for field in weighting_fields:
            w *= self.weighting[field](getattr(gold_ann, field),
                                       getattr(pred_ann, field))
        return w


    def get_matches(self, system, gold):
        """ Assesses the match between sets of annotations

        Returns three lists of items:
        * tp [(item, other_item), ...]
        * fp [(None, other_item), ...]
        * fn [(item, None), ...]
        """
        if self.is_clustering:
            raise ValueError('get_matches is inappropriate '
                             'for {}'.format(self.agg))
        if self.weighting:
            raise NotImplementedError('get_matches not implemented with '
                                      'weighting')
        gold_index = self.build_index(gold)
        pred_index = self.build_index(system)
        gold_keys = keys(gold_index)
        pred_keys = keys(pred_index)
        shared = gold_keys & pred_keys
        tp = [(gold_index[k], pred_index[k]) for k in shared]
        fp = [(None, pred_index[k]) for k in pred_keys - shared]
        fn = [(gold_index[k], None) for k in gold_keys - shared]
        return tp, fp, fn

    def get_overlapping(self, system, gold):
        """
        Returns: overlaps_sys, overlaps_gold
            where each is a dict of sys/gold annotations mapped to a list
            of sorted annotations that overlap and have the same key apart
            from span.
        """
        key_fields = self.key
        if 'span' in key_fields:
            key_fields = [f for f in key_fields if f != 'span'] + ['docid']

        overlaps_sys = {ann: [] for ann in system}
        overlaps_gold = {ann: [] for ann in gold}
        system = self.build_index(system, multi=True, key_fields=key_fields)
        gold = self.build_index(gold, multi=True, key_fields=key_fields)
        for key, sys_annots in system.items():
            gold_annots = gold.pop(key, [])
            sys_annots.sort()
            gold_annots.sort()
            while sys_annots and gold_annots:
                rel = sys_annots[-1].compare_spans(gold_annots[-1])
                if rel != 'non-overlapping':
                    overlaps_sys[sys_annots[-1]].append(gold_annots[-1])
                    overlaps_gold[gold_annots[-1]].append(sys_annots[-1])
                if sys_annots[-1] > gold_annots[-1]:
                    sys_annots.pop()
                else:
                    gold_annots.pop()

        return ({k: list(reversed(v)) for k, v in overlaps_sys.items()},
                {k: list(reversed(v)) for k, v in overlaps_gold.items()})

    @staticmethod
    def measure_overlap(overlaps, mode):
        total = 0.
        if mode == 'sum':
            for ref, cands in overlaps.items():
                if not cands:
                    continue
                # XXX: cands should not be overlapping, but just in case...
                n_chars = 0
                opened = None
                n_open = 0
                offsets = ([(ann.start, '(') for ann in cands] +
                           [(ann.end, ')') for ann in cands])
                offsets.sort()
                for o, d in offsets:  # sorted
                    if d == '(':
                        if opened is None:
                            opened = max(o, ref.start)
                            n_open += 1
                        else:
                            n_open += 1
                    else:
                        assert n_open
                        n_open -= 1
                        if not n_open:
                            n_chars += min(o, ref.end) - opened + 1
                            opened = None
                assert n_open == 0
                assert opened is None
                total += n_chars / (ref.end - ref.start + 1)

        elif mode == 'max':
            for ref, cands in overlaps.items():
                if not cands:
                    continue
                start = ref.start
                end = ref.end
                most = max(min(end, ann.end) - max(start, ann.start)
                           for ann in cands)
                total += (most + 1) / (end - start + 1)
        else:
            raise ValueError('Unexpected overlap measurement mode: %r' % mode)

        return total

    def count_overlap(self, system, gold, gold_mode='sum', sys_mode='sum'):
        # XXX: note by convention modes are gold then sys!
        overlaps_sys, overlaps_gold = self.get_overlapping(system, gold)
        fp = len(overlaps_sys) - self.measure_overlap(overlaps_sys, sys_mode)
        fn = len(overlaps_gold) - self.measure_overlap(overlaps_gold, gold_mode)
        return fp, fn

    def count_clustering(self, system, gold):
        from . import coref_metrics
        if not self.is_clustering:
            raise ValueError('evaluate_clustering is inappropriate '
                             'for {}'.format(self.agg))
        try:
            fn = getattr(coref_metrics, self.agg)
        except AttributeError:
            raise ValueError('Invalid aggregation: {!r}'.format(self.agg))
        if not callable(fn):
            raise ValueError('Invalid aggregation: {!r}'.format(self.agg))
        gold_clusters = self.build_clusters(gold)
        pred_clusters = self.build_clusters(system)
        return fn(gold_clusters, pred_clusters)

    def contingency(self, system, gold):
        if self.is_clustering:
            p_num, p_den, r_num, r_den = self.count_clustering(system, gold)
            ptp = p_num
            fp = p_den - p_num
            rtp = r_num
            fn = r_den - r_num
            return ptp, fp, rtp, fn
        elif self.agg == 'sets':
            tp, fp, fn = self.count_matches(system, gold)
            return tp, fp, tp, fn
        elif self.agg.startswith('overlap-'):
            params = self.agg[len('overlap-'):]
            fp, fn = self.count_overlap(system, gold,
                                        params[:3], params[3:])
            return len(system) - fp, fp, len(gold) - fn, fn
        else:
            # This should not be reachable
            raise ValueError('Unexpected value for agg: %r' % self.agg)



import spacy


class AbbreviationExpander:
    """
    A class to expand abbreviations in text data.

    Args:
    - spacy_model (str): the name of the spaCy model to use. Default is "en_core_sci_sm".

    Attributes:
    - nlp (Language): the spaCy pipeline object.
    - ab (AbbreviationDetector): the abbreviation detector object.
    """

    def __init__(self, spacy_model: str = "en_core_sci_sm"):
        from scispacy.abbreviation import AbbreviationDetector

        self.nlp = spacy.load(spacy_model)
        ab = self.nlp.add_pipe("abbreviation_detector")

    def _expand_abbreviations(self, examples):
        """
        Expands abbreviations in a given example.

        Args:
        - examples (dict): A dictionary with "passages" and "entities" keys containing a list of dictionaries.

        Returns:
        - dict: A dictionary with "entities" key containing a list of dictionaries.
        """

        def get_iter(examples):
            for passages, ents in zip(examples["passages"], examples["entities"]):
                for p in passages:
                    for t, off in zip(p["text"], p["offsets"]):
                        ents_in_text = [
                            e for e in ents if e["offsets"][0][0] >= off[0] and e["offsets"][-1][1] <= off[1]
                        ]
                        yield t, (off, ents_in_text)

        for doc, ctx in self.nlp.pipe(get_iter(examples), as_tuples=True, disable=["ner"]):
            off, ents_in_text = ctx
            abbrv = doc._.abbreviations
            for e in ents_in_text:
                e["long_form"] = None
                e_start, e_end = e["offsets"][0][0], e["offsets"][-1][1]
                for a in abbrv:
                    if e_start - off[0] <= a.start_char and e_end - off[0] >= a.end_char:
                        e["long_form"] = a._.long_form.text
                        abbrv.remove(a)

        return {"entities": examples["entities"]}

    def transform_batch(self, dataset):
        """
        Transforms a given dataset by expanding abbreviations in it.

        Args:
        - dataset (Dataset): A dataset object from the Hugging Face datasets library.

        Returns:
        - Dataset: A transformed dataset object with the same number of examples.
        """
        return dataset.map(self._expand_abbreviations, batched=True)



from tqdm.autonotebook import tqdm
import datasets


def find_context(passages, offsets):
    """
    Finds the left and right context of a target span within a list of passages.

    - passages: a list of passages containing text and offsets
    - offsets: a list of start and end offsets representing the target span

    Returns:
    - The left context target span, and the right context of the target span within the passages.

    Raises:
    - AssertionError if the function fails to find the target span within the passages.
    """
    m_start, m_end = min([o[0] for o in offsets]), max([o[1] for o in offsets])
    for p in passages:
        for t, o in zip(p["text"], p["offsets"]):
            t_start, t_end = o[0], o[1]
            if m_start >= t_start and m_end <= t_end:
                l_context = t[0 : (m_start - t_start)]
                m = t[(m_start - t_start) : (m_end - t_start + 1)]
                r_context = t[(m_end - t_start + 1) :]
                return l_context, m, r_context
    assert False


def get_flat_candidate_ds(candidate_ds, ground_truth, expand_abbreviations, kb):
    """
    Expands the given candidate dataset with additional features, and removes the irrelevant columns.

    Args:
    - candidate_ds: the original dataset containing candidates
    - ground_truth: the ground truth dataset (can be None)
    - expand_abbreviations: a boolean flag indicating whether or not to expand abbreviations
    - kb: the knowledge base

    Returns:
    - A flat candidate dataset with additional columns: "synonyms" and "label".
    - A document index.
    """
    try:
        datasets.disable_progress_bar()
        flat_candidate_ds = candidate_ds.map(
            lambda e, i: get_candidates(e, i, expand_abbreviations),
            batched=True,
            remove_columns=candidate_ds.column_names,
            with_indices=True,
            load_from_cache_file=False,
        )

        doc_index = flat_candidate_ds["doc_index"]
        flat_candidate_ds = flat_candidate_ds.remove_columns(["doc_index"])

        if ground_truth:
            flat_ground_truth = ground_truth.map(
                lambda e, i: get_candidates(e, i, False),
                batched=True,
                remove_columns=ground_truth.column_names,
                with_indices=True,
                load_from_cache_file=False,
            )
            flat_ground_truth = flat_ground_truth.rename_column("candidates", "label")

        synonyms = [
            [[kb.cui_to_entity[cui].canonical_name] + kb.cui_to_entity[cui].aliases for cui in entry]
            for entry in tqdm(flat_candidate_ds["candidates"])
        ]
        semantic_types = [
            [kb.cui_to_entity[cui].types for cui in entry] for entry in tqdm(flat_candidate_ds["candidates"])
        ]
        flat_candidate_ds = flat_candidate_ds.add_column("synonyms", synonyms)
        flat_candidate_ds = flat_candidate_ds.add_column("types", semantic_types)

        if ground_truth:
            flat_candidate_ds = flat_candidate_ds.add_column("label", flat_ground_truth["label"])
    finally:
        datasets.enable_progress_bar()

    return flat_candidate_ds, doc_index


def get_candidates(examples, doc_indices, expand_abbreviations: bool, check_spans=False):
    """
    Retrieves all candidate entities for each mention in the given examples.

    Args:
    - examples: a dataset containing mentions and their corresponding passages
    - doc_indices: a list of indices that correspond to the original documents in the dataset
    - expand_abbreviations: a boolean flag indicating whether or not to expand abbreviations

    Returns:
    - A dictionary containing information about each mention, including its candidates, scores, and context.
    """
    candidates = []
    scores = []
    mentions = []
    doc_ixs = []
    left_contexts = []
    right_contexts = []
    for doc_passages, l, doc_ix in zip(examples["passages"], examples["entities"], doc_indices):
        for e_id, e in enumerate(l):
            doc_ixs.append((doc_ix, e_id))
            context_left, mention, context_right = find_context(doc_passages, e["offsets"])
            if check_spans:
                for mention_text in e["text"]:
                    assert mention_text in mention, (mention_text, mention)
            e_candidates = [n["db_id"] for n in e["normalized"]]
            candidates.append(e_candidates)
            e_scores = [n.get("score", None) for n in e["normalized"]]
            scores.append(e_scores)
            m_text = " ".join(e["text"])
            if expand_abbreviations and e["long_form"]:
                m_text += f' ({e["long_form"]})'
            mentions.append(m_text)
            left_contexts.append(context_left)
            right_contexts.append(context_right)
    return {
        "mention": mentions,
        "candidates": candidates,
        "scores": scores,
        "doc_index": doc_ixs,
        "context_left": left_contexts,
        "context_right": right_contexts,
    }



from datasets import DatasetDict
import numpy as np


class Sampler:
    """
    A class for sampling data from a dataset.

    Args:
    - random_seed (int): The random seed for the numpy random number generator.
    - n (int, optional): The number of samples to extract from the dataset.
    - frac (float, optional): The fraction of the dataset to extract.

    Raises:
        AssertionError: If neither n nor frac is provided.
        AssertionError: If both n and frac are provided.
    """

    def __init__(self, random_seed: int, n: int = None, frac: float = None):
        self.random_seed = random_seed
        self.n = n
        self.frac = frac
        assert n or frac, "n or frac need to be provided"
        assert not (n and frac), "either n or frac can be provided"

    def transform_batch(self, dataset, aligned_dataset=None):
        """
        Samples the dataset or dataset dictionary.

        Args:
        - dataset (Dataset or DatasetDict): The dataset or dataset dictionary to sample.
        - aligned_dataset (DatasetDict, optional): The dataset dictionary to align with `dataset`.

        Returns:
          If `dataset` is a dataset dictionary and `aligned_dataset` is provided, a tuple of two dataset dictionaries
          (one for `dataset` and one for `aligned_dataset`) containing the sampled data. Otherwise, a single
          dataset or dataset dictionary with the sampled data.

        Raises:
        - AssertionError: If neither n nor frac is provided.
        - AssertionError: If both n and frac are provided.
        """

        def _sample(ds, aligned_ds=None):
            n_samples = self.n if self.n else int(self.frac * len(ds))
            rng = np.random.default_rng(seed=52)
            doc_indices = rng.choice(np.arange(0, len(ds)), size=n_samples, replace=False)

            sampled = ds.select(doc_indices)
            if not aligned_ds:
                return sampled
            return sampled, aligned_ds.select(doc_indices)

        if type(dataset) == DatasetDict:
            res = DatasetDict()
            if aligned_dataset:
                aligned_res = DatasetDict()
            for k, v in dataset.items():
                if aligned_dataset:
                    res[k], aligned_res[k] = _sample(v, aligned_dataset[k])
                else:
                    res[k] = _sample(v)
            return (res, aligned_res) if aligned_dataset else res
        else:
            return _sample(dataset, aligned_dataset or None)



from .data_loader import (
    QueryDataset, 
    QueryDataset_custom, 
    QueryDataset_COMETA,
    QueryDataset_pretraining, 
    MetricLearningDataset,
    MetricLearningDataset_pairwise,
    DictionaryDataset,
)

from .metric_learning import Sap_Metric_Learning
from ....linkers.model_wrapper import Model_Wrapper



from xmen.umls import get_semantic_groups


class SemanticGroupFilter:
    """
    A class for filtering entities in a dataset based on their semantic group.

    Args:
    - kb (EntityKnowledgeBase): The knowledge base containing the entities and their types.
    - version (str, optional): The version of the semantic group mapping to use.
    """

    def __init__(self, kb, version="v03"):
        self._kb = kb
        self._sem_groups = get_semantic_groups(version).set_index("TUI")

    def __getstate__(self):
        return {}

    def get_sem_groups(self, cui):
        """
        Returns the semantic group(s) of an entity.

        Args:
        - cui (str): The concept unique identifier of the entity.

        Returns:
        - A list of the semantic group(s) of the entity, or an empty list if the entity is not found in the knowledge base.
        """
        entry = self._kb.cui_to_entity.get(cui, None)
        if not entry:
            return []
        return [self._sem_groups.loc[t].GRP for t in entry.types]

    def filter_semantic_groups(self, example):
        """
        Filters the entities in an example based on their semantic group.

        Args:
        - example (dict): A dictionary containing the example entities.

        Returns:
        - A dictionary containing the filtered entities.
        """
        entities = example["entities"]
        for e in entities:
            sem_groups = e["type"]
            if not isinstance(sem_groups, list):
                sem_groups = [sem_groups]
            filtered = []
            for n in e["normalized"]:
                concept_groups = self.get_sem_groups(n["db_id"])
                if any([g for g in sem_groups if g in concept_groups]):
                    filtered.append(n)
            e["normalized"] = filtered
        return {"entities": entities}

    def transform_batch(self, ds):
        """
        Applies the semantic group filter to a dataset.

        Args:
        - ds (Dataset): The dataset to apply the filter to.

        Returns:
        - The filtered dataset.
        """
        return ds.map(lambda e: self.filter_semantic_groups(e), load_from_cache_file=False)



# pylint: disable=g-import-not-at-top,g-bad-import-order,wrong-import-position

from abc import ABC, abstractmethod
from datasets.arrow_dataset import Dataset
from pathlib import Path
from xmen.log import logger

from xmen.reranking import Reranker


class EntityLinker(ABC):
    def predict_batch(self, dataset: Dataset, batch_size: int = None) -> Dataset:
        """Naive default implementation of batch prediction.
        Should be overridden if the particular model provides an efficient way to predict in batch (e.g., on a GPU)

        Args:
            dataset (Dataset): Input (arrow) dataset with entities but without concepts

        Returns:
            Dataset: (arrow) dataset with linked concepts
        """
        return dataset.map(
            lambda unit: {"entities": self.predict(unit["passages"], unit["entities"])},
            load_from_cache_file=False,
        )

    def get_logger(self):
        return logger

    @abstractmethod
    def predict(self, passages: list, entities: list) -> list:
        pass


class RerankedLinker(EntityLinker):
    def __init__(self, linker: EntityLinker, ranker: Reranker):
        self.linker = linker
        self.ranker = ranker

    def predict_batch(self, dataset: Dataset, **linker_kwargs) -> Dataset:
        result = self.linker.predict_batch(dataset, linker_kwargs)
        return self.ranker.rerank_batch(result)

    def predict(self, passages: list, entities: list) -> list:
        raise NotImplementedError()


from .tf_idf_ngram_linker import TFIDFNGramLinker
from .sap_bert_linker import SapBERTLinker
from .ensemble import EnsembleLinker


def default_ensemble(index_base_path, k_ngram=100, k_sapbert=1000, cuda=True):
    """
    Creates the default ensemble candidate generator consisting of equally weighted SapBERT and TF-IDF N-Gram Linker
    """
    index_base_path = Path(index_base_path)
    ngram_linker = TFIDFNGramLinker(index_base_path=index_base_path / "ngrams", k=k_ngram)

    SapBERTLinker.clear()
    sapbert_linker = SapBERTLinker(index_base_path=index_base_path / "sapbert", k=k_sapbert, cuda=cuda)

    ensemble = EnsembleLinker()
    ensemble.add_linker("ngram", ngram_linker, k=k_ngram)
    ensemble.add_linker("sapbert", sapbert_linker, k=k_sapbert)

    return ensemble



#!/usr/bin/env python
"""
Evaluate linker performance.
Adapted from https://github.com/wikilinks/neleval/blob/master/neleval/annotation.py for Python >3.10 compatibility
"""
import warnings
import json
import itertools
from collections import OrderedDict, defaultdict

from .configs import (DEFAULT_MEASURE_SET, parse_measures,
                      MEASURE_HELP, get_measure, load_weighting)
from .document import Document, Reader
from neleval.utils import log, utf8_open, json_dumps


class StrictMetricWarning(Warning):
    pass


METRICS = [
    'ptp',
    'fp',
    'rtp',
    'fn',
    'precision',
    'recall',
    'fscore',
]


class Evaluate(object):
    'Evaluate system output'

    def __init__(self, system, gold=None,
                 measures=DEFAULT_MEASURE_SET,
                 fmt='none', group_by=None, overall=False,
                 type_weights=None):
        """
        system - system output
        gold - gold standard
        measures - measure definitions to use
        fmt - output format
        """
        if not isinstance(system, list):
            log.debug('Reading system output..')
            system = list(Reader(utf8_open(system)))
            log.debug('..done.')
        if not isinstance(gold, list):
            log.debug('Reading gold standard..')
            gold = list(Reader(utf8_open(gold)))
            log.debug('..done.')
        self.system = system
        self.gold = gold
        self.measures = parse_measures(measures or DEFAULT_MEASURE_SET)
        self.format = self.FMTS[fmt] if not callable(fmt) else fmt
        self.doc_pairs = list(self.iter_pairs(self.system, self.gold))
        self.group_by = group_by
        self.overall = overall
        self.weighting = load_weighting(type_weights=type_weights)

    @classmethod
    def iter_pairs(self, system, gold):
        sdocs = {d.id: d for d in system}
        gdocs = {d.id: d for d in gold}
        for docid in set(sdocs.keys()).union(gdocs.keys()):
            sdoc = sdocs.get(docid) or Document(docid, [])
            gdoc = gdocs.get(docid) or Document(docid, [])
            yield sdoc, gdoc

    def __call__(self, measures=None):
        measures = (parse_measures(measures)
                    if measures is not None
                    else self.measures)
        self.results = OrderedDict()

        # XXX: should avoid grouping by doc in the first place
        system_annotations = [ann for doc in self.system
                              for ann in doc.annotations]
        gold_annotations = [ann for doc in self.gold
                            for ann in doc.annotations]
        if not self.group_by:
            name_fmt = '{measure}'
            system_grouped = {((),): system_annotations}
            gold_grouped = {((),): gold_annotations}
        else:
            name_fmt = '{measure}'
            for i, field in enumerate(self.group_by):
                name_fmt += ';{}={{group[{}]}}'.format(field, i)
            get_group = lambda ann: tuple(getattr(ann, field)
                                          for field in self.group_by)

            system_grouped = defaultdict(list)
            for ann in system_annotations:
                system_grouped[get_group(ann)].append(ann)

            gold_grouped = defaultdict(list)
            for ann in gold_annotations:
                gold_grouped[get_group(ann)].append(ann)

        if self.group_by:
            n_fields = len(self.group_by)
            group_vals = [sorted(set(group[i] for group in gold_grouped))
                          for i in range(n_fields)]
        else:
            group_vals = [((),)]

        for measure in measures:
            measure_mats = []
            for group in itertools.product(*group_vals):
                # XXX should we only be accounting for groups that are non-empty in gold? non empty in either sys or gold?
                mat = Matrix(*get_measure(measure, weighting=self.weighting).
                             contingency(system_grouped[group],
                                         gold_grouped[group])
                             )
                measure_mats.append((group, mat))

                if not self.group_by or not self.overall:
                    name = name_fmt.format(measure=measure,
                                           group=[json.dumps(v) for v in group])
                    self.results[name] = mat.results

            if self.group_by:
                # Macro-averages for each field
                micro_labels = ['<micro>'] * len(self.group_by)
                for i in range(n_fields):
                    constituents = defaultdict(Matrix)
                    for group, mat in measure_mats:
                        constituents[group[i]] += mat

                    labels = micro_labels[:]
                    labels[i] = '<macro>'
                    name = name_fmt.format(measure=measure, group=labels)
                    self.results[name] = macro_average(constituents.values())

                # Overall micro-average
                name = name_fmt.format(measure=measure, group=micro_labels)
                self.results[name] = sum(constituents.values(),
                                         Matrix()).results

        return self.format(self, self.results)

    @classmethod
    def add_arguments(cls, p):
        p.add_argument('system', metavar='FILE')
        p.add_argument('-g', '--gold', required=True)
        p.add_argument('-f', '--fmt', default='tab', choices=cls.FMTS.keys())
        p.add_argument('-m', '--measure', dest='measures', action='append',
                       metavar='NAME', help=MEASURE_HELP)
        p.add_argument('-b', '--group-by', dest='group_by', action='append',
                       metavar='FIELD',
                       help='Report results per field-value, '
                            'and micro/macro-averaged over these, '
                            'Multiple --group-by may be used.  '
                            'E.g. -b docid -b type.  '
                            'NB: micro-average may not equal overall score.')
        p.add_argument('--by-doc', dest='group_by', action='append_const',
                       const='docid', help='Alias for -b docid')
        p.add_argument('--by-type', dest='group_by', action='append_const',
                       const='type', help='Alias for -b type')
        p.add_argument('--overall', default=False, action='store_true',
                       help='With --group-by, report only overall, not per-group results')
        p.add_argument('--type-weights', metavar='FILE', default=None,
                       help='File mapping gold and sys types to a weight, '
                       'such as produced by weights-for-hierarchy')
        p.set_defaults(cls=cls)
        return p

    @classmethod
    def count_all(cls, doc_pairs, measures, weighting=None):
        for m in measures:
            yield (m,) + cls.count(m, doc_pairs, weighting=weighting)

    @classmethod
    def count(cls, measure, doc_pairs, weighting=None):
        per_doc = []
        measure = get_measure(measure, weighting=weighting)
        for sdoc, gdoc in doc_pairs:
            per_doc.append(Matrix(*measure.contingency(sdoc.annotations,
                                                       gdoc.annotations)))
        overall = sum(per_doc, Matrix())
        return per_doc, overall

    # formatters

    def tab_format(self, results, num_fmt='{:.3f}', delimiter='\t'):
        lines = [self._header(delimiter)]
        for measure, measure_results in sorted(results.items()):
            row = self.row(results, measure, num_fmt)
            lines.append(delimiter.join(row))
        return '\n'.join(lines)

    @staticmethod
    def _header(delimiter='\t'):
        return delimiter.join([i[:6] for i in METRICS] + ['measure'])

    def row(self, results, measure_str, num_fmt):
        row = []
        measure_results = results.get(measure_str, {})
        for metric in METRICS:
            val = measure_results.get(metric, 0)
            if isinstance(val, float):
                row.append(num_fmt.format(val))
            else:
                row.append(str(val))
        row.append(measure_str)
        return row

    @classmethod
    def read_tab_format(cls, file):
        header = next(file)
        assert header.rstrip() == cls._header(), 'Differing headers: expected {!r}, got {!r}'.format(cls._header(), header.rstrip())
        results = {}
        for l in file:
            row = l.rstrip().split('\t')
            measure = row.pop()
            row = map(float, row)
            results[measure] = dict(zip(METRICS, row))
        return results

    def json_format(self, results):
        return json_dumps(results)

    def no_format(self, results):
        return results

    FMTS = {
        'tab': tab_format,
        'json': json_format,
        'none': no_format,
    }


class Matrix(object):
    def __init__(self, ptp=0, fp=0, rtp=0, fn=0):
        self.ptp = ptp
        self.fp = fp
        self.rtp = rtp
        self.fn = fn

    def __str__(self):
        return 'ptp={},fp={},rtp={},fn={}'.format(self.ptp,
                                                  self.fp,
                                                  self.rtp,
                                                  self.fn)

    def __add__(self, other):
        return Matrix(self.ptp + other.ptp,
                      self.fp + other.fp,
                      self.rtp + other.rtp,
                      self.fn + other.fn)

    def __iadd__(self, other):
        self.ptp += other.ptp
        self.fp += other.fp
        self.rtp += other.rtp
        self.fn += other.fn
        return self

    @property
    def results(self):
        return {
            'precision': self.precision,
            'recall': self.recall,
            'fscore': self.fscore,
            'ptp': self.ptp,
            'fp': self.fp,
            'rtp': self.rtp,
            'fn': self.fn,
            }

    @property
    def precision(self):
        return self.div(self.ptp, self.ptp+self.fp)

    @property
    def recall(self):
        return self.div(self.rtp, self.rtp+self.fn)

    def div(self, n, d):
        if d == 0:
            warnings.warn('Strict P/R defaulting to zero score for '
                          'zero denominator',
                          StrictMetricWarning)
            return 0.0
        else:
            return n / float(d)

    @property
    def fscore(self):
        p = self.precision
        r = self.recall
        return self.div(2*p*r, p+r)


def macro_average(results_iter):
    out = defaultdict(float)
    for i, results in enumerate(results_iter):
        if hasattr(results, 'results'):
            # accept Matrix instances instead
            results = results.results
        for k, v in results.items():
            out[k] += v
    return {k: v / (i + 1) for k, v in out.items()}



from collections.abc import Mapping
from typing import List
from xmen.reranking import Reranker


class RuleBasedReranker(Reranker):
    """
    Reranks a dataset using a list of rules.

    Args:
    - rules (List[Mapping]): A list of mappings representing the rules to be applied.

    Attributes:
    - rules (List[Mapping]): A list of mappings representing the rules to be applied.

    Returns:
    - The reranked dataset.
    """

    def __init__(self, rules: List[Mapping]) -> None:
        self.rules = rules

    def rerank_batch(self, dataset):
        def apply_rules(concepts, context):
            for r in self.rules:
                concepts = r(concepts, context)
            return concepts

        def apply_rules_unit(entry):
            mapped_entities = []
            for e in entry["entities"]:
                e = e.copy()
                concepts = e["normalized"]
                concepts = apply_rules(concepts, e)
                e.update({"normalized": concepts})
                mapped_entities.append(e)
            return {"entities": mapped_entities}

        return dataset.map(apply_rules_unit, load_from_cache_file=False)



from typing import Dict
from xmen.linkers import EntityLinker
from xmen.data import filter_and_apply_threshold, features, util
from datasets import Dataset, utils, DatasetDict
import itertools


class EnsembleLinker(EntityLinker):
    """
    A class for combining multiple entity linking models together into an ensemble.

    Attributes:
    - linkers_fn (Dict[str, Callable]): a dictionary of linker functions, each associated with a unique name
    - linkers_k (Dict[str, int]): a dictionary of integers representing the number of top entities to keep for each linker
    - linker_thresholds (Dict[str, float]): a dictionary of floats representing the score threshold to use for each linker
    - linker_weigths (Dict[str, float]): a dictionary of floats representing the weight to give to each linker during combination
    """

    # Ignore caching when using dataset.map
    def __getstate__(self):
        return {}

    def __init__(self):
        self.linkers_fn = {}
        self.linkers_k = {}
        self.linker_thresholds = {}
        self.linker_weigths = {}

    def add_linker(self, name, linker, k=1, threshold=0.0, weight=1.0):
        """
        Adds a new linker to the ensemble.

        Args:
        - name: a unique name to associate with the linker
        - linker: an instance of the EntityLinker class
        - k: an integer representing the number of top entities to keep for this linker (default is 1)
        - threshold: a float representing the score threshold to use for this linker (default is 0.0)
        - weight: a float representing the weight to give to this linker during combination (default is 1.0)
        """
        self.add_linker_fn(name, lambda: linker, k, threshold, weight)

    def add_linker_fn(self, name, linker_fn, k=1, threshold=0.0, weight=1.0):
        """
        Adds a new linker function to the ensemble.

        Args:
        - name: a unique name to associate with the linker
        - linker_fn: a function that returns an instance of the EntityLinker class
        - k: an integer representing the number of top entities to keep for this linker (default is 1)
        - threshold: a float representing the score threshold to use for this linker (default is 0.0)
        - weight: a float representing the weight to give to this linker during combination (default is 1.0)
        """
        self.linkers_fn[name] = linker_fn
        self.linkers_k[name] = k
        self.linker_thresholds[name] = threshold
        self.linker_weigths[name] = weight

    def predict_batch(self, dataset, batch_size=1, top_k=None, reuse_preds=None):
        """
        Runs the ensemble on a dataset in batches.

        Args:
        - dataset: a dataset of documents to run entity linking on
        - batch_size: an integer representing the batch size to use
        - top_k: an optional integer representing the number of top entities to keep per document
        - reuse_preds: an optional dictionary of precomputed predictions for each linker

        Returns:
        - a dictionary containing the predicted entities for each document in the dataset
        """

        def merge_linkers(batch, index, reuse_preds_split=None):
            progress = utils.logging.is_progress_bar_enabled()
            try:
                batch = util.init_schema(batch)
                mapped = {}
                if progress:
                    utils.logging.disable_progress_bar()

                for linker_name, linker_fn in self.linkers_fn.items():
                    if reuse_preds_split:
                        linked = reuse_preds_split[linker_name].select(index)
                    else:
                        linker = linker_fn()
                        linked = linker.predict_batch(Dataset.from_dict(batch, features=features), batch_size)
                    mapped[linker_name] = filter_and_apply_threshold(
                        linked,
                        self.linkers_k[linker_name],
                        self.linker_thresholds[linker_name],
                    )["entities"]

                entities = []

                for i, doc in enumerate(batch["entities"]):
                    for j, e in enumerate(doc):
                        merged_scores = []
                        e["normalized"] = []
                        for linker_name in mapped.keys():
                            for n in mapped[linker_name][i][j]["normalized"]:
                                n["predicted_by"] = linker_name
                                if "score" in n:
                                    n["score"] *= self.linker_weigths[linker_name]
                                merged_scores.append(n)
                        # Merge duplicate predictions
                        key_fn = lambda cand: cand["db_id"]
                        for key, grp in itertools.groupby(sorted(merged_scores, key=key_fn), key=key_fn):
                            preds = list(grp)
                            db_name = {p["db_name"] for p in preds}
                            score = max([p["score"] for p in preds])
                            pred_by = [p["predicted_by"] for p in preds]
                            assert len(db_name) == 1
                            e["normalized"].append(
                                {
                                    "db_id": key,
                                    "db_name": db_name.pop(),
                                    "score": score,
                                    "predicted_by": pred_by,
                                }
                            )
                        e["normalized"] = sorted(e["normalized"], key=lambda n: n["score"])[-1::-1]
                        if top_k:
                            e["normalized"] = e["normalized"][:top_k]

                    entities.append(doc)
            finally:
                if progress:
                    utils.logging.enable_progress_bar()
            return {"entities": entities}

        if type(dataset) == DatasetDict:
            return DatasetDict(
                {
                    split: dataset[split].map(
                        lambda b, i: merge_linkers(
                            b, i, {linker: v[split] for linker, v in reuse_preds.items()} if reuse_preds else None
                        ),
                        with_indices=True,
                        batched=True,
                        batch_size=batch_size,
                        load_from_cache_file=False,
                        features=features,
                    )
                    for split in dataset.keys()
                }
            )
        else:
            return dataset.map(
                lambda b, i: merge_linkers(b, i, reuse_preds),
                with_indices=True,
                batched=True,
                batch_size=batch_size,
                load_from_cache_file=False,
                features=features,
            )

    def predict(self, unit: str, entities: dict) -> dict:
        raise NotImplementedError()



import torch
import numpy as np

from dataclasses import dataclass
from transformers.tokenization_utils_base import (
    PreTrainedTokenizerBase,
    PaddingStrategy,
)
from transformers import EvalPrediction

from typing import Optional, Union, Dict


def simple_accuracy(preds, labels):
    """
    Computes the simple accuracy of predictions.

    Args:
    - preds (Union[List, np.ndarray]): List or array of predicted labels.
    - labels (Union[List, np.ndarray]): List or array of true labels.

    Returns:
    - float: Simple accuracy of predictions.
    """
    return (preds == labels).mean()


def top_1_metrics(preds, labels):
    """
    Computes top-1 metrics for a given set of predictions.

    Args:
    - preds (np.ndarray): Array of predicted labels.
    - labels (np.ndarray): Array of true labels.

    Returns:
    - Dict[str, float]: A dictionary containing the following metrics:
        - pr_nil: Precision for predicting NIL entities.
        - re_nil: Recall for predicting NIL entities.
        - f1_nil: F1 score for predicting NIL entities.
        - pr: Precision for predicting all entities.
        - re: Recall for predicting all entities.
        - f1: F1 score for predicting all entities.

    """
    tp_nil = tp = tp_no_nil = fp = fn_nil = fn = fn_no_nil = fp_no_nil = 0
    for p, l in zip(preds, labels):
        if p[0] == 0 and l == 0:
            tp_nil += 1
        elif p[0] == l:
            tp += 1
        elif p[0] == 0:
            if p[1] == l and l != 0:
                tp_no_nil += 1
            else:
                fn_no_nil += 1
                fp_no_nil += 1
            fn_nil += 1
        elif p[0] != 0:
            fn += 1
            fp += 1
    f1 = lambda x, y: 2 * x * y / (x + y) if x + y > 0 else 0

    pr_nil = tp / (tp + fp) if tp > 0 else 0
    re_nil = tp / (tp + fn + fn_nil + tp_nil) if tp > 0 else 0

    pr = (tp + tp_no_nil) / (tp + tp_no_nil + fp + fp_no_nil + tp_nil)
    re = (tp + tp_no_nil) / (tp + tp_no_nil + fn + fn_no_nil + tp_nil)

    return {
        "pr_nil": pr_nil,
        "re_nil": re_nil,
        "f1_nil": f1(pr_nil, re_nil),
        "pr": pr,
        "re": re,
        "f1": f1(pr, re),
    }


def compute_entity_linking_metrics(p: EvalPrediction) -> Dict:
    """
    Computes entity linking metrics for a given set of predictions.

    Args:
    - p (EvalPrediction): Object containing predictions and labels.

    Returns:
    - Dict[str, float]: A dictionary containing the following metrics:
        - pr_nil: Precision for predicting NIL entities.
        - re_nil: Recall for predicting NIL entities.
        - f1_nil: F1 score for predicting NIL entities.
        - pr: Precision for predicting all entities.
        - re: Recall for predicting all entities.
        - f1: F1 score for predicting all entities.
        - acc: Simple accuracy of predictions.
    """
    preds = np.argsort(p.predictions, axis=1)[:, -1::-1]

    metrics = top_1_metrics(preds, p.label_ids)
    metrics["acc"] = simple_accuracy(preds[:, 0], p.label_ids)
    return metrics


@dataclass
class DataCollatorForMultipleChoice:
    """
    Data collator that will dynamically pad the inputs for multiple choice received.

    Args:
    - tokenizer (:obj:`PreTrainedTokenizerBase`): The tokenizer used for encoding the data.
    - padding (:obj:`Union[bool, str, PaddingStrategy]`, optional, defaults to True):
         Select a padding strategy. Default strategy is to pad to the longest sample in the batch.
    - max_length (:obj:`int`, optional):
         The maximum length of the tokenized input sequences. Will truncate all samples longer than this.
    - pad_to_multiple_of (:obj:`int`, optional):
         If set, the input sequences will be padded to a length that is a multiple of the given value.

    Returns:
    - A dictionary with the following key-value pairs:
        - input_ids (:obj:`torch.Tensor`): The input token ids.
        - attention_mask (:obj:`torch.Tensor`): The input attention mask.
        - token_type_ids (:obj:`torch.Tensor`, optional): The token type
    """

    tokenizer: PreTrainedTokenizerBase
    padding: Union[bool, str, PaddingStrategy] = True
    max_length: Optional[int] = None
    pad_to_multiple_of: Optional[int] = None

    def __call__(self, features):
        label_name = "label" if "label" in features[0].keys() else "labels"
        labels = [feature.pop(label_name) for feature in features]
        batch_size = len(features)
        num_choices = len(features[0]["input_ids"])
        flattened_features = [
            [{k: v[i] for k, v in feature.items()} for i in range(num_choices)] for feature in features
        ]
        flattened_features = sum(flattened_features, [])

        batch = self.tokenizer.pad(
            flattened_features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )

        batch = {k: v.view(batch_size, num_choices, -1) for k, v in batch.items()}
        input_ids = batch["input_ids"]

        batch["labels"] = torch.tensor(labels, dtype=torch.int64)
        return batch
